import os
import pytest

import numpy as np
import pandas as pd

from lingam import RESIT
from lingam.utils import visualize_nonlinear_causal_effect

from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
from sklearn.svm import LinearSVC


@pytest.fixture
def test_data_adj():
    np.random.seed(0)

    X = _generate_data()

    reg = RandomForestRegressor()
    model = RESIT(regressor=reg, alpha=1)
    model.fit(X)

    adj_mat = model.adjacency_matrix_
    cause_name = "X2"
    effect_name = "X3"
    return X, adj_mat, cause_name, effect_name

@pytest.fixture
def test_data_bs():
    np.random.seed(0)

    X = _generate_data()

    reg = RandomForestRegressor()
    model = RESIT(regressor=reg, alpha=1)

    bs_result = model.bootstrap(X, n_sampling=3)
    cause_name = "X2"
    effect_name = "X3"
    return X, bs_result, cause_name, effect_name

def _generate_data(n_samples=100):
    def N1(size):
        return np.random.uniform(size=size) - 0.5

    X = np.zeros([n_samples, 5])
    X[:, 0] = N1(n_samples)
    X[:, 1] = 3 * (X[:, 0] + 0.25) * (X[:, 0] - 0.25) + N1(n_samples)
    X[:, 2] = -0.75 * (X[:, 0] - 1) * (X[:, 0] - 2) + 1.5 * X[:, 1] + N1(n_samples)
    X[:, 3] = 5 * (X[:, 1] + 0.4) * (X[:, 1] - 0.1) * (X[:, 1] - 0.5) + 1 * np.log(5 * X[:, 2] + 20) + N1(n_samples)
    X[:, 4] = -0.8 * (X[:, 3] - 1.5) * (X[:, 3] - 3.5) + N1(n_samples)

    X = pd.DataFrame(X, columns=[f"X{i}" for i in range(5)])
    return X

def test_exec_sucess_adj(test_data_adj):
    X, adj_mat, cause_name, effect_name = test_data_adj

    # continuous
    fig = visualize_nonlinear_causal_effect(
        X,
        adj_mat,
        RandomForestRegressor(),
        cause_name,
        effect_name
    )
    assert len(fig.axes) == 1

    # discrete
    X2 = _to_discrete(X, effect_name)
    fig = visualize_nonlinear_causal_effect(
        X2,
        adj_mat,
        RandomForestClassifier(),
        cause_name,
        effect_name
    )
    assert len(fig.axes) == 3

def test_exec_sucess_bs(test_data_bs):
    X, bs_result, cause_name, effect_name = test_data_bs

    # continuous
    fig = visualize_nonlinear_causal_effect(
        X,
        bs_result,
        RandomForestRegressor(),
        cause_name,
        effect_name
    )
    assert len(fig.axes) == 1

    # discrete
    X2 = _to_discrete(X, effect_name)
    fig = visualize_nonlinear_causal_effect(
        X2,
        bs_result,
        RandomForestClassifier(),
        cause_name,
        effect_name
    )
    assert len(fig.axes) == 3

def _to_discrete(X, name, q=(33, 66)):
    X = X.copy()
    delims = np.percentile(X[name], q=q)
    X[name] = X[name].map(lambda x: sum([delim <= x for delim in delims])) 
    X[name] = pd.Categorical(X[name])
    return X

def test_exec_exceptioin_args_check(test_data_adj):
    X, adj_mat, cause_name, effect_name = test_data_adj
    X2 = _to_discrete(X, effect_name)

    # X
    try:
        fig = visualize_nonlinear_causal_effect(
            np.array(X),
            adj_mat,
            RandomForestRegressor(),
            cause_name,
            effect_name
        )
    except TypeError:
        pass
    else:
        raise AssertionError

    # cd_result
    try:
        fig = visualize_nonlinear_causal_effect(
            X,
            adj_mat.reshape(-1, 1),
            RandomForestRegressor(),
            cause_name,
            effect_name
        )
    except TypeError:
        pass
    else:
        raise AssertionError

    try:
        fig = visualize_nonlinear_causal_effect(
            X,
            adj_mat.astype(str),
            RandomForestRegressor(),
            cause_name,
            effect_name
        )
    except ValueError:
        pass
    else:
        raise AssertionError

    # cause_name
    try:
        fig = visualize_nonlinear_causal_effect(
            X,
            adj_mat,
            RandomForestRegressor(),
            "cause_name",
            effect_name
        )
    except ValueError:
        pass
    else:
        raise AssertionError

    # effect_ame
    try:
        fig = visualize_nonlinear_causal_effect(
            X,
            adj_mat,
            RandomForestRegressor(),
            cause_name,
            "effect_name"
        )
    except ValueError:
        pass
    else:
        raise AssertionError

    # estmator
    try:
        fig = visualize_nonlinear_causal_effect(
            X2,
            adj_mat,
            RandomForestRegressor(),
            cause_name,
            effect_name
        )
    except TypeError:
        pass
    else:
        raise AssertionError

    try:
        fig = visualize_nonlinear_causal_effect(
            X,
            adj_mat,
            RandomForestClassifier(),
            cause_name,
            effect_name
        )
    except TypeError:
        pass
    else:
        raise AssertionError

    try:
        fig = visualize_nonlinear_causal_effect(
            X2,
            adj_mat,
            LinearSVC(),
            cause_name,
            effect_name
        )
    except RuntimeError:
        pass
    else:
        raise AssertionError

    # cause_positions
    try:
        fig = visualize_nonlinear_causal_effect(
            X,
            adj_mat,
            RandomForestRegressor(),
            cause_name,
            effect_name,
            cause_positions=[1, 2, 3, None]
        )
    except TypeError:
        pass
    else:
        raise AssertionError

    # percentile
    try:
        fig = visualize_nonlinear_causal_effect(
            X,
            adj_mat,
            RandomForestRegressor(),
            cause_name,
            effect_name,
            percentile=(0, 30, 60, 100)
        )
    except TypeError:
        pass
    else:
        raise AssertionError

    try:
        fig = visualize_nonlinear_causal_effect(
            X,
            adj_mat,
            RandomForestRegressor(),
            cause_name,
            effect_name,
            percentile=(0, 50, 101)
        )
    except TypeError:
        pass
    else:
        raise AssertionError

    # fig
    try:
        fig = visualize_nonlinear_causal_effect(
            X,
            adj_mat,
            RandomForestRegressor(),
            cause_name,
            effect_name,
            fig = 0
        )
    except TypeError:
        pass
    else:
        raise AssertionError




import json
import pytest

import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.dummy import DummyRegressor, DummyClassifier
from sklearn.svm import SVR, SVC
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import GridSearchCV
from scipy.special import expit

import os, sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from lingam.causal_based_simulator import CausalBasedSimulator, TrainResult

DATA_DIR_PATH = os.path.dirname(__file__) + "/test_causal_based_simulator"

ENABLE_CAPTURE = False
if ENABLE_CAPTURE:
    if not os.path.isdir(DATA_DIR_PATH):
        os.mkdir(DATA_DIR_PATH)

@pytest.fixture
def init():
    return lambda :np.random.seed(0)

@pytest.fixture
def test_data():
    np.random.seed(0)

    N = 1000

    G = np.array([[0.0, 0.0, 0.0, 3.0, 0.0, 0.0],
                  [3.0, 0.0, 2.0, 0.0, 0.0, 0.0],
                  [0.0, 0.0, 0.0, 6.0, 0.0, 0.0],
                  [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                  [8.0, 0.0,-1.0, 0.0, 0.0, 0.0],
                  [4.0, 0.0, 0.0, 0.0, 0.0, 0.0]])
    X = {}
    X["x3"] = np.random.uniform(size=N)
    X["x2"] = G[2, 3] * X["x3"] + np.random.uniform(size=N)
    X["x0"] = G[0, 3] * X["x3"] + np.random.uniform(size=N)
    X["x4"] = G[4, 2] * X["x2"] + G[4, 0] * X["x0"] + np.random.uniform(size=N)
    X["x1"] = G[1, 2] * X["x2"] + G[1, 0] * X["x0"] + np.random.uniform(size=N)
    X["x5"] = G[5, 0] * X["x0"] + np.random.uniform(size=N)
    X = dict(sorted(X.items()))
    X = pd.DataFrame(X)

    return X, G

@pytest.fixture
def test_data2():
    np.random.seed(0)

    N = 1000

    G = np.array([[0.0, 0.0, 0.0, 3.0, 0.0, 0.0],
                  [3.0, 0.0, 1.0, 0.0, 0.0, 0.0],
                  [0.0, 0.0, 0.0, 1.5, 0.0, 0.0],
                  [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                  [8.0, 0.0, 1.0, 0.0, 0.0, 0.0],
                  [4.0, 0.0, 0.0, 0.0, 0.0, 0.0]])
    
    X = {}
    
    X["x3"] = np.random.uniform(size=N)
    
    X["x2"] = np.empty(N).astype(str)
    for i in range(N):
        prob_a = expit(G[2, 3] * X["x3"][i] - 0.5)
        X["x2"][i] = np.random.choice(["a", "b"], p=[prob_a, 1 - prob_a])
    
    X["x0"] = G[0, 3] * X["x3"] + np.random.uniform(size=N)
    
    X["x4"] = np.empty(N)
    for i in range(N):
        if X["x2"][i] == "a":
            X["x4"][i] = 8.0 * X["x0"][i] + np.random.uniform()
        elif X["x2"][i] == "b":
            X["x4"][i] = 4.0 * X["x0"][i] + np.random.uniform()
    
    X["x1"] = np.empty(N)
    for i in range(N):
        if X["x2"][i] == "a":
            X["x1"][i] = 3.0 * X["x0"][i] + np.random.uniform()
        elif X["x2"][i] == "b":
            X["x1"][i] = 1.5 * X["x0"][i] + np.random.uniform()
    
    X["x5"] = G[5, 0] * X["x0"] + np.random.uniform(size=1000)
    
    X = dict(sorted(X.items()))
    X = pd.DataFrame(X)
    X["x2"] = X["x2"].astype("category")

    return X, G

@pytest.fixture
def test_data3():
    np.random.seed(0)

    G = np.array([[0.0, 0.0, 0.0],
                  [2.0, 0.0, 0.0],
                  [2.0, 0.0, 0.0]])

    N = 1000
    
    X = {}
    
    X["x0"] = np.random.choice(["a", "b"], p=[0.75, 0.25], size=N)
    
    X["x1"] = []
    for i in range(N):
        if X["x0"][i] == "a":
            X["x1"].append(np.random.uniform() * 2)
        elif X["x0"][i] == "b":
            X["x1"].append(np.random.uniform())
            
    X["x2"] = []
    for i in range(N):
        if X["x0"][i] == "a":
            X["x2"].append(np.random.choice(["c", "d"], p=[0.75, 0.25]))
        elif X["x0"][i] == "b":
            X["x2"].append(np.random.choice(["c", "d"], p=[0.25, 0.75]))
            
            
    X = pd.DataFrame(X)
    X["x0"] = X["x0"].astype("category")
    X["x2"] = X["x2"].astype("category")
    X = X[sorted(X.columns)]

    return X, G

def test_check_data(init, test_data, test_data2):
    init()

    X, causal_graph = test_data
    X2, causal_graph2 = test_data2

    # Normal
    sim = CausalBasedSimulator()
    try:
        X_, categorical_info = sim._check_data(X)
    except RuntimeError:
        raise AssertionError
    else:
        pass

    # Normal2: categorical variable
    sim = CausalBasedSimulator()
    try:
        X2_, categorical_info = sim._check_data(X2)
    except RuntimeError:
        raise AssertionError
    else:
        pass

    # Exception: wrog type
    X3 = np.ones((10, 5))
    sim = CausalBasedSimulator()
    try:
        X3_, categorical_info = sim._check_data(X3)
    except RuntimeError:
        pass
    else:
        raise AssertionError

    # Exception2: wrong type
    X4 = np.ones((10, 5)).tolist()
    sim = CausalBasedSimulator()
    try:
        X4_, categorical_info = sim._check_data(X4)
    except RuntimeError:
        pass
    else:
        raise AssertionError

def test_check_causal_graph(init, test_data):
    init()

    X, causal_graph = test_data

    # Normal
    sim = CausalBasedSimulator()
    try:
        causal_graph = sim._check_causal_graph(causal_graph, 6)
    except RuntimeError:
        raise AssertionError
    else:
        pass

    # Exception: wrong type
    sim = CausalBasedSimulator()
    try:
        causal_graph_ = sim._check_causal_graph("not array", 3)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Exception: wrong shape
    causal_graph2 = np.ones((3, 4))
    sim = CausalBasedSimulator()
    try:
        causal_graph2_ = sim._check_causal_graph(causal_graph2, 3)
    except RuntimeError:
        pass
    else:
        raise AssertionError

def test_check_models(init, test_data2):
    init()

    X, causal_graph = test_data2
    categorical_info = {"x2":["a", "b"]}

    # Normal
    sim = CausalBasedSimulator()
    models = {
        "x0": SVR(kernel="linear"),
        "x2": SVC(kernel="linear", probability=True),
        "x4": SVR(kernel="linear"),
    }
    try:
        sim._check_models(models, X.index, X.columns, categorical_info)
    except:
        raise AssertionError
    else:
        pass

    # Normal: Pipeline
    sim = CausalBasedSimulator()
    models = {
        "x0": Pipeline([
            ("preprocess", StandardScaler()),
            ("estimator", SVR(kernel="linear")),
        ]),
        "x2": SVC(kernel="linear", probability=True),
        "x4": SVR(kernel="linear"),
    }
    try:
        sim._check_models(models, X.index, X.columns, categorical_info)
    except:
        raise AssertionError
    else:
        pass

    # Normal: GridSearchCV
    sim = CausalBasedSimulator()
    models = {
        "x0": GridSearchCV(SVR(kernel="linear"), {"C": [0.1**i for i in range(3)]}),
        "x2": SVC(kernel="linear", probability=True),
        "x4": SVR(kernel="linear"),
    }
    try:
        sim._check_models(models, X.index, X.columns, categorical_info)
    except:
        raise AssertionError
    else:
        pass

    # Exception: wrong type
    sim = CausalBasedSimulator()
    models = (
        ("x0", SVR(kernel="linear")),
        ("x2", SVC(kernel="linear", probability=True)),
        ("x4", SVR(kernel="linear")),
    )
    try:
        sim._check_models(models, X.index, X.columns, categorical_info)
    except RuntimeError:
        pass
    else:
        raise AssertionError

    # Exception: wrong task type
    sim = CausalBasedSimulator()
    models = {
        "x0": SVC(kernel="linear", probability=True),
        "x2": SVC(kernel="linear", probability=True),
        "x4": SVR(kernel="linear"),
    }
    try:
        sim._check_models(models, X.index, X.columns, categorical_info)
    except RuntimeError:
        pass
    else:
        raise AssertionError


    # Exception: non existent variable name
    sim = CausalBasedSimulator()
    models = {
        "x0": SVR(kernel="linear"),
        "x2": SVC(kernel="linear", probability=True),
        "x6": SVR(kernel="linear"),
    }
    try:
        sim._check_models(models, X.index, X.columns, categorical_info)
    except RuntimeError:
        pass
    else:
        raise AssertionError

    # Exception: not instance
    sim = CausalBasedSimulator()
    models = {
        "x0": SVR(kernel="linear"),
        "x2": LogisticRegression,
        "x4": SVR(kernel="linear"),
    }
    try:
        sim._check_models(models, X.index, X.columns, categorical_info)
    except RuntimeError:
        pass
    else:
        raise AssertionError

    # Exception: Pipeline, wrong task
    sim = CausalBasedSimulator()
    models = {
        "x0": Pipeline([
            ("preprocess", StandardScaler()),
            ("estimator", SVC(kernel="linear", probability=True)),
        ]),
        "x2": SVC(kernel="linear", probability=True),
        "x4": SVR(kernel="linear"),
    }
    try:
        sim._check_models(models, X.index, X.columns, categorical_info)
    except:
        pass
    else:
        raise AssertionError

    # Exception: GridSearchCV, wrong task
    sim = CausalBasedSimulator()
    models = {
        "x0": GridSearchCV(SVC(kernel="linear", probability=True), {"C": [0.1**i for i in range(3)]}),
        "x2": SVC(kernel="linear", probability=True),
        "x4": SVR(kernel="linear"),
    }
    try:
        sim._check_models(models, X.index, X.columns, categorical_info)
    except:
        pass
    else:
        raise AssertionError

def test_changing_exog(init, test_data2):
    init()

    X, causal_graph = test_data2
    categorical_info = {"x2": ["a", "b"]}

    # Normal
    sim = CausalBasedSimulator()
    changing_exog = {
        "x0": np.random.uniform(size=X.shape[0]),
        "x1": np.random.uniform(size=X.shape[0]),
        "x3": np.random.uniform(size=X.shape[0]),
    }
    try:
        sim._check_changing_exog(changing_exog, X.index, X.columns, categorical_info)
    except RuntimeError:
        raise AssertionError
    else:
        pass

    # Exception: wrong type
    sim = CausalBasedSimulator()
    changing_exog = (
        ("x0", np.random.uniform(size=X.shape[0])),
        ("x1", np.random.uniform(size=X.shape[0])),
        ("x3", np.random.uniform(size=X.shape[0])),
    )
    try:
        sim._check_changing_exog(changing_exog, X.index, X.columns, categorical_info)
    except RuntimeError:
        pass
    else:
        raise AssertionError

    # Exception: non existent variable name
    sim = CausalBasedSimulator()
    changing_exog = {
        "x0": np.random.uniform(size=X.shape[0]),
        "x1": np.random.uniform(size=X.shape[0]),
        "x6": np.random.uniform(size=X.shape[0]),
    }
    try:
        sim._check_changing_exog(changing_exog, X.index, X.columns, categorical_info)
    except RuntimeError:
        pass
    else:
        raise AssertionError

    # Exception: can't specify to categorical variable
    sim = CausalBasedSimulator()
    changing_exog = {
        "x2": np.random.choice(["a", "b"], size=X.shape[0]),
    }
    try:
        sim._check_changing_exog(changing_exog, X.index, X.columns, categorical_info)
    except RuntimeError:
        pass
    else:
        raise AssertionError

    # Exception: wrong shape
    sim = CausalBasedSimulator()
    changing_exog = {
        "x0": np.random.uniform(size=X.shape[0]),
        "x1": np.random.uniform(size=X.shape[0] + 1),
        "x3": np.random.uniform(size=X.shape[0]),
    }
    try:
        sim._check_changing_exog(changing_exog, X.index, X.columns, categorical_info)
    except RuntimeError:
        pass
    else:
        raise AssertionError

def test_changing_models(init, test_data2):
    init()

    X, causal_graph = test_data2
    categorical_info = {"x2": ["a", "b"]}
    X["x2"] = X["x2"].apply(lambda x: categorical_info["x2"].index(x)).astype(int)
    train_results = {
        "x0": [
            TrainResult(condition=None, exp_columns=["x3"], model=None, predicted=None, residual=None),
        ],
        "x1": [
            TrainResult(condition={"x2": "a"}, exp_columns=["x0"], model=None, predicted=None, residual=None),
            TrainResult(condition={"x2": "b"}, exp_columns=["x0"], model=None, predicted=None, residual=None)
        ],
        "x2": [
            TrainResult(condition=None, exp_columns=["x3"], model=None, predicted=None, residual=None),
        ],
        "x3": [],
        "x4": [],
        "x5": [],
    }
    x0_model = LinearRegression()
    x0_model.fit(X[train_results["x0"][0].exp_columns].values, X["x0"] * 10)
    X_a = X[X["x2"] == categorical_info["x2"].index("a")]
    x1_model_a = LinearRegression()
    x1_model_a.fit(X_a[train_results["x1"][0].exp_columns].values, X_a["x1"] * 10)
    X_b = X[X["x2"] == categorical_info["x2"].index("b")]
    x1_model_b = LinearRegression()
    x1_model_b.fit(X_b[train_results["x1"][1].exp_columns].values, X_b["x1"] * 10)
    x2_model = LogisticRegression()
    x2_model.fit(X[train_results["x2"][0].exp_columns].values, X["x2"] * 10)
    x0_pipe = Pipeline([("estimator", LinearRegression())])
    x0_pipe.fit(X[train_results["x0"][0].exp_columns].values, X["x0"] * 10)
    x0_gs = GridSearchCV(SVR(kernel="linear"), {"C": [0.1**i for i in range(3)]})
    x0_gs.fit(X[train_results["x0"][0].exp_columns].values, X["x0"] * 10)

    # Normal
    sim = CausalBasedSimulator()
    changing_models = [
        {
            "name": "x0",
            "condition": None,
            "model": x0_model
        },
        {
            "name": "x1",
            "condition": {"x2": "a"},
            "model": x1_model_a
        },
        {
            "name": "x1",
            "condition": {"x2": "b"},
            "model": x1_model_b
        },
        {
            "name": "x2",
            "condition": None,
            "model": x2_model
        }
    ]
    try:
        sim._check_changing_models(changing_models, categorical_info, train_results)
    except RuntimeError:
        raise AssertionError
    else:
        pass

    # Normal: Pipeline
    sim = CausalBasedSimulator()
    changing_models = [
        {
            "name": "x0",
            "condition": None,
            "model": x0_pipe
        },
    ]
    try:
        sim._check_changing_models(changing_models, categorical_info, train_results)
    except RuntimeError:
        raise AssertionError
    else:
        pass

    # Normal: GridSearchCV
    sim = CausalBasedSimulator()
    changing_models = [
        {
            "name": "x0",
            "condition": None,
            "model": x0_gs
        },
    ]
    try:
        sim._check_changing_models(changing_models, categorical_info, train_results)
    except RuntimeError:
        raise AssertionError
    else:
        pass

    # Exception: type
    sim = CausalBasedSimulator()
    changing_models = {
        "x0": {
            "name": "x0",
            "condition": None,
            "model": x0_model
        },
    }
    try:
        sim._check_changing_models(changing_models, categorical_info, train_results)
    except RuntimeError:
        pass
    else:
        raise AssertionError

    # Exception: list of element
    sim = CausalBasedSimulator()
    changing_models = [
        "x0"
    ]
    try:
        sim._check_changing_models(changing_models, categorical_info, train_results)
    except RuntimeError:
        pass
    else:
        raise AssertionError

    # Exception: lack of key
    sim = CausalBasedSimulator()
    changing_models = [
        {
            "condition": None,
            "model": x0_model
        },
 
    ]
    try:
        sim._check_changing_models(changing_models, categorical_info, train_results)
    except RuntimeError:
        pass
    else:
        raise AssertionError

    # Exception: lask of key
    sim = CausalBasedSimulator()
    changing_models = [
        {
            "name": "x0",
            "model": x0_model
        },
 
    ]
    try:
        sim._check_changing_models(changing_models, categorical_info, train_results)
    except RuntimeError:
        pass
    else:
        raise AssertionError

    # Exception: lackk of key
    sim = CausalBasedSimulator()
    changing_models = [
        {
            "name": "x0",
            "condition": None,
        },
 
    ]
    try:
        sim._check_changing_models(changing_models, categorical_info, train_results)
    except RuntimeError:
        pass
    else:
        raise AssertionError

    # Exception: non existent variable name
    sim = CausalBasedSimulator()
    changing_models = [
        {
            "name": "x6",
            "condition": None,
            "model": x0_model
        },
    ]
    try:
        sim._check_changing_models(changing_models, categorical_info, train_results)
    except RuntimeError:
        pass
    else:
        raise AssertionError

    # Exception: on existent condition
    sim = CausalBasedSimulator()
    changing_models = [
        {
            "name": "x1",
            "condition": {"x2":"c"},
            "model": x1_model_a
        },
    ]
    try:
        sim._check_changing_models(changing_models, categorical_info, train_results)
    except RuntimeError:
        pass
    else:
        raise AssertionError

    # Exception: model is None
    sim = CausalBasedSimulator()
    changing_models = [
        {
            "name": "x1",
            "condition": None,
            "model": None
        },
    ]
    try:
        sim._check_changing_models(changing_models, categorical_info, train_results)
    except RuntimeError:
        pass
    else:
        raise AssertionError

    # Exception: wrong task type
    sim = CausalBasedSimulator()
    changing_models = [
        {
            "name": "x1",
            "condition": None,
            "model": x2_model
        },
    ]
    try:
        sim._check_changing_models(changing_models, categorical_info, train_results)
    except RuntimeError:
        pass
    else:
        raise AssertionError

    # Exception: wrong model type
    sim = CausalBasedSimulator()
    changing_models = [
        {
            "name": "x0",
            "condition": None,
            "model": x2_model
        },
    ]
    try:
        sim._check_changing_models(changing_models, categorical_info, train_results)
    except RuntimeError:
        pass
    else:
        raise AssertionError

def test_get_causal_order(init, test_data):
    init()
    X, causal_graph = test_data

    # Normal
    sim = CausalBasedSimulator()
    ret = sim._get_causal_order(causal_graph)
    assert ret.tolist() == [3, 0, 2, 1, 4, 5]

def _extract_model_params(train_result):
    model_params = {c: {} for c in train_result.keys()}

    for col, conds in train_result.items():
        for cond in conds:
            condition = str(cond["condition"])

            if hasattr(cond["model"], "intercept_"):
                params = {
                    "coef": cond["model"].coef_.tolist(),
                    "intercept": cond["model"].intercept_.tolist(),
                }
            elif isinstance(cond["model"], DummyRegressor):
                params = {
                    "expected_value": cond["model"].constant_.tolist(),
                }
            elif isinstance(cond["model"], DummyClassifier):
                params = {
                    "classes": cond["model"].classes_.tolist(),
                    "p": cond["model"].class_prior_.tolist(),
                }
            else:
                raise AssertionError

            model_params[col][condition] = params
            
    return model_params

def _read_train_result_attrs(name, sim):
    get_fname = lambda x: f"{DATA_DIR_PATH}/{name}_{x}.json"
    fname_resid = get_fname("resid")
    fname_params = get_fname("params")

    if ENABLE_CAPTURE:
        residual = sim.residual_
        with open(fname_resid, "w") as f:
            r = residual.to_dict("list")
            f.write(json.dumps(r))

        model_params = _extract_model_params(sim.train_result_)
        with open(fname_params, "w") as f:
            f.write(json.dumps(model_params))

    with open(fname_resid, "r") as f:
        residual = pd.DataFrame(json.loads(f.read()))

    with open(fname_params, "r") as f:
        model_params = json.loads(f.read())

    return residual, model_params

def _is_same_data(df, df2, exclude_cols=[], nan_cols=[]):
    if df.shape != df2.shape:
        raise AssertionError

    if (df.index != df2.index).any():
        raise AssertionError

    if (df.columns != df2.columns).any():
        raise AssertionError

    if (df.dtypes != df2.dtypes).any():
        raise AssertionError

    # numeric
    non_cat = df.dtypes != "category"
    non_cat2 = df2.dtypes != "category"
    if list(non_cat) != list(non_cat2):
        raise AssertionError

    removes = exclude_cols + nan_cols
    for c in removes:
        non_cat[list(df.columns).index(c)] = False
        non_cat2[list(df2.columns).index(c)] = False
    assert (np.isclose(df.loc[:, non_cat], df2.loc[:, non_cat2])).all()

    # categorical
    cat = df.dtypes == "category"
    cat2 = df2.dtypes == "category"
    if list(cat) != list(cat2):
        raise AssertionError
    if len(cat) == 0 and len(cat2) == 0:
        assert np.all(df.loc[:, cat] == df2.loc[:, cat2])

def _is_same_coef_intercept(model_params, model_params2):
    if len(set(model_params.keys()) - set(model_params2)) != 0:
        raise AssertionError

    for column in model_params.keys():
        for cond, result in model_params[column].items():
            result2 = model_params[column][cond]
            coef, intercept = result["coef"], result["intercept"]
            coef2, intercept2 = result2["coef"], result2["intercept"]

            assert np.isclose(coef, coef2).all()
            assert np.isclose(intercept, intercept2).all()

def test_train(init, test_data, test_data2, test_data3):
    init()

    X, causal_graph = test_data
    X2, causal_graph2 = test_data2
    X3, causal_graph3 = test_data3

    # Normal
    sim = CausalBasedSimulator()
    sim.train(X, causal_graph)

    residual, model_params = _read_train_result_attrs("test_train", sim)
    model_params2 = _extract_model_params(sim.train_result_)

    _is_same_data(residual, sim.residual_, nan_cols=["x3"])
    _is_same_coef_intercept(model_params, model_params2)

    # Normal2
    sim = CausalBasedSimulator()
    models = {
        "x0": SVR(kernel="linear"),
        "x2": SVR(kernel="linear"),
        "x4": SVR(kernel="linear"),
    }
    sim.train(X, causal_graph, models=models)

    residual, model_params = _read_train_result_attrs("test_train2", sim)
    model_params2 = _extract_model_params(sim.train_result_)

    _is_same_data(residual, sim.residual_, nan_cols=["x3"])
    _is_same_coef_intercept(model_params, model_params2)

    # Normal3
    sim = CausalBasedSimulator()
    sim.train(X2, causal_graph2)

    residual, model_params = _read_train_result_attrs("test_train3", sim)
    model_params2 = _extract_model_params(sim.train_result_)

    _is_same_data(residual, sim.residual_, nan_cols=["x2", "x3"])
    _is_same_coef_intercept(model_params, model_params2)

    # Normal4
    sim = CausalBasedSimulator()
    sim.train(X3, causal_graph3)

    residual, model_params = _read_train_result_attrs("test_train4", sim)
    model_params2 = _extract_model_params(sim.train_result_)

    _is_same_data(residual, sim.residual_, nan_cols=["x0", "x2"])
    for cond in model_params2["x1"]:
        assert np.isclose(model_params["x1"][cond]["expected_value"], model_params2["x1"][cond]["expected_value"])
    for cond in model_params2["x2"]:
        assert np.all(model_params["x2"][cond]["classes"] == model_params2["x2"][cond]["classes"])
        assert np.all(np.isclose(model_params["x2"][cond]["p"], model_params2["x2"][cond]["p"]))

def test_run(init, test_data, test_data2, test_data3):
    init()

    X, causal_graph = test_data
    X2, causal_graph2 = test_data2
    X3, causal_graph3 = test_data3

    # Normal: no optios
    sim = CausalBasedSimulator()
    sim.train(X, causal_graph)
    sim.run()

    if ENABLE_CAPTURE:
        sim.simulated_data_.to_csv(f"{DATA_DIR_PATH}/test_run-simulated_data.csv", index=None)

    truth_df = pd.read_csv(f"{DATA_DIR_PATH}/test_run-simulated_data.csv")
    _is_same_data(sim.simulated_data_, truth_df)

    # Normal: only exogeous
    sim = CausalBasedSimulator()
    sim.train(X, causal_graph)
    changing_exog = {
        "x0": 2 * X["x0"].copy(),
    }
    sim.run(changing_exog=changing_exog)

    if ENABLE_CAPTURE:
        sim.simulated_data_.to_csv(f"{DATA_DIR_PATH}/test_run2-simulated_data.csv", index=None)

    truth_df = pd.read_csv(f"{DATA_DIR_PATH}/test_run2-simulated_data.csv")
    _is_same_data(sim.simulated_data_, truth_df)

    # Normal: only models
    sim = CausalBasedSimulator()
    sim.train(X, causal_graph)
    train_x = X[sim.train_result_["x2"][0]["exp_columns"]].copy()
    train_x["x3"] /= 4
    train_y = X["x2"]
    model = LinearRegression()
    model.fit(train_x.values, train_y.values)
    changing_models = [
        {
            "name": "x2",
            "condition": None,
            "model": model
        }
    ]
    sim.run(changing_models=changing_models)

    if ENABLE_CAPTURE:
        sim.simulated_data_.to_csv(f"{DATA_DIR_PATH}/test_run3-simulated_data.csv", index=None)

    truth_df = pd.read_csv(f"{DATA_DIR_PATH}/test_run3-simulated_data.csv")
    _is_same_data(sim.simulated_data_, truth_df)

    # Normal: exogenous and models
    sim = CausalBasedSimulator()
    sim.train(X, causal_graph)
    exp_columns_x2 = sim.train_result_["x2"][0]["exp_columns"]
    train_x = X[exp_columns_x2].copy()
    train_x["x3"] /= 4
    train_y = X["x2"]
    model = LinearRegression()
    model.fit(train_x.values, train_y.values)
    changing_models = [
        {
            "name": "x2",
            "condition": None,
            "model": model
        }
    ]
    changing_exog = {
        "x2": train_y - model.predict(X[exp_columns_x2].values)
    }
    sim.run(changing_exog=changing_exog, changing_models=changing_models)

    if ENABLE_CAPTURE:
        sim.simulated_data_.to_csv(f"{DATA_DIR_PATH}/test_run4-simulated_data.csv", index=None)

    truth_df = pd.read_csv(f"{DATA_DIR_PATH}/test_run4-simulated_data.csv")
    _is_same_data(sim.simulated_data_, truth_df)

    # Normal: no options
    sim = CausalBasedSimulator()
    sim.train(X2, causal_graph2)
    sim.run()

    if ENABLE_CAPTURE:
        sim.simulated_data_.to_csv(f"{DATA_DIR_PATH}/test_run5-simulated_data.csv", index=None)

    truth_df = pd.read_csv(f"{DATA_DIR_PATH}/test_run5-simulated_data.csv")
    truth_df["x2"] = truth_df["x2"].astype("category")
    _is_same_data(sim.simulated_data_, truth_df)

    # Normal: no optios
    sim = CausalBasedSimulator()
    sim.train(X3, causal_graph3)
    sim.run()

    if ENABLE_CAPTURE:
        sim.simulated_data_.to_csv(f"{DATA_DIR_PATH}/test_run6-simulated_data.csv", index=None)

    truth_df = pd.read_csv(f"{DATA_DIR_PATH}/test_run6-simulated_data.csv")
    truth_df["x0"] = truth_df["x0"].astype("category")
    truth_df["x2"] = truth_df["x2"].astype("category")
    _is_same_data(sim.simulated_data_, truth_df)



import random

import numpy as np
import pandas as pd

from lingam.utils import (
    print_causal_directions,
    print_dagc,
    make_prior_knowledge,
    remove_effect,
    make_dot,
    make_dot_highlight,
    predict_adaptive_lasso,
    get_sink_variables,
    get_exo_variables,
    find_all_paths,
    extract_ancestors,
    f_correlation,
)


def test_print_causal_directions():
    cdc = {"from": [0, 1, 2, 3], "to": [1, 2, 3, 4], "count": [5, 4, 3, 2]}
    print_causal_directions(cdc, n_sampling=5)
    print_causal_directions(cdc, n_sampling=5, labels=["A", "B", "C", "D", "E"])


def test_print_dagc():
    dagc = {
        "dag": [
            {"from": [0, 1, 2, 3], "to": [1, 2, 3, 4]},
            {"from": [0, 1, 2, 3], "to": [1, 2, 3, 0]},
            {"from": [0, 1, 2, 3], "to": [1, 2, 3, 4]},
        ],
        "count": [5, 2, 1],
    }
    print_dagc(dagc, n_sampling=5)
    print_dagc(dagc, n_sampling=5, labels=["A", "B", "C", "D", "E"])


def test_make_prior_knowledge():
    pk = make_prior_knowledge(
        n_variables=5,
        exogenous_variables=[0],
        sink_variables=[1],
        paths=[[2, 3]],
        no_paths=[[3, 4]],
    )
    assert pk.shape[0] == 5 and pk.shape[1] == 5


def test_get_sink_variables():
    am = [
        [0, 1, 1, 0],
        [0, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 1, 0, 0],
    ]
    sv = get_sink_variables(adjacency_matrix=am)
    assert len(sv) == 2


def test_get_exo_variables():
    am = [
        [0, 1, 1, 0],
        [0, 0, 0, 0],
        [0, 0, 0, 0],
        [0, 1, 0, 0],
    ]
    ev = get_exo_variables(adjacency_matrix=am)
    assert len(ev) == 2


def test_remove_effect():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0 * x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=["x0", "x1", "x2", "x3"])

    remove_effect(X, remove_features=[0])


def test_make_dot():
    # default
    am = [
        [0, 1, 1, 0],
        [0, 0, 0, 0],
        [0, 0, 0, 0],
        [0, 1, 0, 0],
    ]
    ev = make_dot(am, labels=["x0", "x1", "x2", "x3"])

    # default
    am = [
        [0, 1, 1, 0],
        [0, 0, np.nan, 0],
        [0, np.nan, 0, 0],
        [0, 1, 0, 0],
    ]
    ev = make_dot(am, labels=["x0", "x1", "x2", "x3"])

    am = [
        [0, 0, 0, 1, 0, 0],
        [1, 0, 1, 0, 0, 0],
        [0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [1, 0, 1, 0, 0, 0],
        [1, 0, 0, 0, 0, 0],
    ]
    ev = make_dot(am, path=(3, 1), path_color="red", detect_cycle=True)

    # prediction
    ev = make_dot(am, prediction_feature_indices=[1, 2], prediction_coefs=[0.1, 0.1])
    ev = make_dot(
        am,
        prediction_feature_indices=[1, 2],
        prediction_target_label="Target",
        prediction_line_color="#0000FF",
    )
    ev = make_dot(
        am, prediction_feature_indices=[1, 2], prediction_feature_importance=[0.5, 0.2]
    )

    # invalid adjacency matrix
    am = [
        [0, 1, 1, 0],
        [0, 0, 0, 0],
        [0, 0, 0, 0],
    ]
    try:
        ev = make_dot(am, labels=["x0", "x1", "x2", "x3"])
    except ValueError:
        pass
    else:
        raise AssertionError

    # invalid label size
    am = [
        [0, 1, 1, 0],
        [0, 0, 0, 0],
        [0, 0, 0, 0],
        [0, 1, 0, 0],
    ]
    try:
        ev = make_dot(am, labels=["x0", "x1", "x2"])
    except ValueError:
        pass
    else:
        raise AssertionError

    # invalid predict settings
    am = [
        [0, 1, 1, 0],
        [0, 0, 0, 0],
        [0, 0, 0, 0],
        [0, 1, 0, 0],
    ]
    try:
        ev = make_dot(am, prediction_feature_indices=[1, 2], prediction_coefs=[0.1])
    except ValueError:
        pass
    else:
        raise AssertionError

    try:
        ev = make_dot(
            am, prediction_feature_indices=[1, 2], prediction_feature_importance=[0.5]
        )
    except ValueError:
        pass
    else:
        raise AssertionError


def test_make_dot_highlight():
    # default
    am = [
        [0, 0, 0, 1, 0, 0],
        [1, 0, 1, 0, 0, 0],
        [0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [1, 0, 1, 0, 0, 0],
        [1, 0, 0, 0, 0, 0],
    ]
    d = make_dot_highlight(am, 0)
    d = make_dot_highlight(am, 0, labels=["x0", "x1", "x2", "x3", "x4", "x5"])
    d = make_dot_highlight(am, 0, max_dsc=0, max_anc=None)
    d = make_dot_highlight(am, 0, max_dsc=0, max_anc=None, draw_others=False)
    d = make_dot_highlight(am, 0, detect_cycle=True)
    d = make_dot_highlight(am, 0, detect_cycle=True, lower_limit=0.1)
    d = make_dot_highlight(am, 0, lower_limit=0.001, cmap="cool", vmargin=3, hmargin=3)


def test_predict_adaptive_lasso():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0 * x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=["x0", "x1", "x2", "x3"])

    predict_adaptive_lasso(X.values, predictors=[0, 1, 2], target=3)


def test_find_all_paths():
    dag = [
        [0, 1, 1, 0],
        [0, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 1, 0, 0],
    ]
    find_all_paths(np.array(dag), 1, 0)

    # min_causal_effect
    dag = [
        [0, 2, 1, 0],
        [0, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 1, 0, 0],
    ]
    find_all_paths(np.array(dag), 1, 0, min_causal_effect=1)

    # Invalid DAG: cycle
    dag = [
        [1, 1, 1, 1],
        [1, 1, 1, 1],
        [1, 1, 1, 1],
        [1, 1, 1, 1],
    ]
    try:
        find_all_paths(np.array(dag), 1, 0)
    except ValueError:
        pass
    else:
        raise AssertionError


def test_extract_ancestors():
    def get_coef():
        coef = random.random()
        return coef if coef >= 0.5 else coef - 1.0

    get_external_effect = lambda n: np.random.normal(0.0, 0.5, n) ** 3

    B = np.array(
        [
            [0.0, 0.0, get_coef(), 0.0],
            [0.0, 0.0, 0.0, 0.0],
            [0.0, get_coef(), 0.0, 0.0],
            [0.0, 0.0, get_coef(), 0.0],
        ]
    )

    samples = 1000
    x1 = get_external_effect(samples)
    x2 = x1 * B[2, 1] + get_external_effect(samples)
    x0 = x2 * B[0, 2] + get_external_effect(samples)
    x3 = x2 * B[3, 2] + get_external_effect(samples)

    # x5, x6 is a latent variable.
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=["x0", "x1", "x2", "x3"])

    # Extract a set of ancestors of each variable
    extract_ancestors(X)


def corr_data(n, r):
    a = np.random.randn(n)
    e1 = np.random.randn(n)
    e2 = np.random.randn(n)
    if r < 0:
        r = -r
        x = -np.sqrt(r) * a - np.sqrt(1 - r) * e1
    else:
        x = np.sqrt(r) * a + np.sqrt(1 - r) * e1
    y = np.sqrt(r) * a + np.sqrt(1 - r) * e2
    return x, y


def test_f_correlation():
    # size < 1000
    x, y = corr_data(500, r=0.3)
    f_correlation(x, y)

    # size > 1000
    x, y = corr_data(1500, r=0.3)
    f_correlation(x, y)

    # len(x) != len(y)
    try:
        x = np.array([1, 2, 3])
        y = np.array([1, 2, 3, 4])
        f_correlation(x, y)
    except ValueError:
        pass
    else:
        raise AssertionError



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

import itertools
import warnings
from abc import ABCMeta, abstractmethod

import numpy as np
from sklearn.utils import check_array

from .bootstrap import BootstrapMixin
from .hsic import hsic_test_gamma
from .utils import predict_adaptive_lasso


class _BaseLiNGAM(BootstrapMixin, metaclass=ABCMeta):
    """Base class for all LiNGAM algorithms."""

    def __init__(self, random_state=None):
        """Construct a _BaseLiNGAM model.

        Parameters
        ----------
        random_state : int, optional (default=None)
            random_state is the seed used by the random number generator.
        """
        self._random_state = random_state
        self._causal_order = None
        self._adjacency_matrix = None

    @abstractmethod
    def fit(self, X):
        """Subclasses should implement this method!
        Fit the model to X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        self : object
            Returns the instance itself.
        """

    def estimate_total_effect(self, X, from_index, to_index):
        """Estimate total effect using causal model.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Original data, where n_samples is the number of samples
            and n_features is the number of features.
        from_index :
            Index of source variable to estimate total effect.
        to_index :
            Index of destination variable to estimate total effect.

        Returns
        -------
        total_effect : float
            Estimated total effect.
        """
        # Check parameters
        X = check_array(X)

        # Check from/to causal order
        from_order = self._causal_order.index(from_index)
        to_order = self._causal_order.index(to_index)
        if from_order > to_order:
            warnings.warn(
                f"The estimated causal effect may be incorrect because "
                f"the causal order of the destination variable (to_index={to_index}) "
                f"is earlier than the source variable (from_index={from_index})."
            )

        # from_index + parents indices
        parents = np.where(np.abs(self._adjacency_matrix[from_index]) > 0)[0]
        predictors = [from_index]
        predictors.extend(parents)

        # Estimate total effect
        coefs = predict_adaptive_lasso(X, predictors, to_index)

        return coefs[0]

    def get_error_independence_p_values(self, X):
        """Calculate the p-value matrix of independence between error variables.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Original data, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        independence_p_values : array-like, shape (n_features, n_features)
            p-value matrix of independence between error variables.
        """
        # Check parameters
        X = check_array(X)
        n_samples = X.shape[0]
        n_features = X.shape[1]

        E = X - np.dot(self._adjacency_matrix, X.T).T
        p_values = np.zeros([n_features, n_features])
        for i, j in itertools.combinations(range(n_features), 2):
            _, p_value = hsic_test_gamma(
                np.reshape(E[:, i], [n_samples, 1]), np.reshape(E[:, j], [n_samples, 1])
            )
            p_values[i, j] = p_value
            p_values[j, i] = p_value

        return p_values

    def _estimate_adjacency_matrix(self, X, prior_knowledge=None):
        """Estimate adjacency matrix by causal order.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        prior_knowledge : array-like, shape (n_variables, n_variables), optional (default=None)
            Prior knowledge matrix.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        if prior_knowledge is not None:
            pk = prior_knowledge.copy()
            np.fill_diagonal(pk, 0)

        B = np.zeros([X.shape[1], X.shape[1]], dtype="float64")
        for i in range(1, len(self._causal_order)):
            target = self._causal_order[i]
            predictors = self._causal_order[:i]

            # Exclude variables specified in no_path with prior knowledge
            if prior_knowledge is not None:
                predictors = [p for p in predictors if pk[target, p] != 0]

            # target is exogenous variables if predictors are empty
            if len(predictors) == 0:
                continue

            B[target, predictors] = predict_adaptive_lasso(X, predictors, target)

        self._adjacency_matrix = B
        return self

    @property
    def causal_order_(self):
        """Estimated causal ordering.

        Returns
        -------
        causal_order_ : array-like, shape (n_features)
            The causal order of fitted model, where
            n_features is the number of features.
        """
        return self._causal_order

    @property
    def adjacency_matrix_(self):
        """Estimated adjacency matrix.

        Returns
        -------
        adjacency_matrix_ : array-like, shape (n_features, n_features)
            The adjacency matrix B of fitted model, where
            n_features is the number of features.
        """
        return self._adjacency_matrix



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

import numpy as np
from sklearn.utils import check_array

from .base import _BaseLiNGAM
from .hsic import hsic_test_gamma


class RESIT(_BaseLiNGAM):
    """Implementation of RESIT(regression with subsequent independence test) Algorithm [1]_

    References
    ----------
    .. [1] Jonas Peters, Joris M Mooij, Dominik Janzing, and Bernhard Sch ̈olkopf.
       Causal discovery with continuous additive noise models.
       Journal of Machine Learning Research, 15:2009-2053, 2014.

    Notes
    -----
    RESIT algorithm returns an **adjacency matrix consisting of zeros or ones**,
    rather than an adjacency matrix consisting of causal coefficients,
    in order to estimate nonlinear causality.
    """

    def __init__(self, regressor, random_state=None, alpha=0.01):
        """Construct a RESIT model.

        Parameters
        ----------
        regressor : regressor object implementing 'fit' and 'predict' function (default=None)
            Regressor to compute residuals.
            This regressor object must have ``fit`` method and ``predict`` function like scikit-learn's model.
        random_state : int, optional (default=None)
            ``random_state`` is the seed used by the random number generator.
        alpha : float, optional (default=0.01)
            Alpha level for HSIC independence test when removing superfluous edges.
        """
        # Check parameters
        if regressor is None:
            raise ValueError("Specify regression model in 'regressor'.")
        else:
            if not (hasattr(regressor, "fit") and hasattr(regressor, "predict")):
                raise ValueError("'regressor' has no fit or predict method.")

        if alpha < 0.0:
            raise ValueError("alpha must be an float greater than 0.")

        super().__init__(random_state)
        self._alpha = alpha
        self._reg = regressor

    def fit(self, X):
        """Fit the model to X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        # Check parameters
        X = check_array(X)
        n_features = X.shape[1]

        # Determine topological order
        pa, pi = self._estimate_order(X)

        # Remove superfluous edges
        pa = self._remove_edges(X, pa, pi)

        # Create adjacency matrix from parent-child relationship
        adjacency_matrix = np.zeros([n_features, n_features])
        for i, parents in pa.items():
            for p in parents:
                adjacency_matrix[i, p] = 1

        self._causal_order = pi
        self._adjacency_matrix = adjacency_matrix
        return self

    def _estimate_order(self, X):
        """Determine topological order"""
        S = np.arange(X.shape[1])
        pa = {}
        pi = []
        for _ in range(X.shape[1]):

            if len(S) == 1:
                pa[S[0]] = []
                pi.insert(0, S[0])
                continue

            hsic_stats = []
            for k in S:
                # Regress Xk on {Xi}
                predictors = [i for i in S if i != k]
                self._reg.fit(X[:, predictors], X[:, k])
                residual = X[:, k] - self._reg.predict(X[:, predictors])
                # Measure dependence between residuals and {Xi}
                hsic_stat, hsic_p = hsic_test_gamma(residual, X[:, predictors])
                hsic_stats.append(hsic_stat)

            k = S[np.argmin(hsic_stats)]
            S = S[S != k]
            pa[k] = S.tolist()
            pi.insert(0, k)

        return pa, pi

    def _remove_edges(self, X, pa, pi):
        """Remove superfluous edges"""
        for k in range(1, X.shape[1]):
            parents = pa[pi[k]].copy()

            for l in parents:
                # Regress Xk on {Xi}
                predictors = [i for i in pa[pi[k]] if i != l]
                # if len(predictors) <= 1:
                if len(predictors) < 1:
                    continue
                self._reg.fit(X[:, predictors], X[:, pi[k]])
                residual = X[:, pi[k]] - self._reg.predict(X[:, predictors])
                # Measure dependence between residuals and {Xi}
                _, hsic_p = hsic_test_gamma(residual, X[:, predictors])
                # eliminate edge
                if hsic_p > self._alpha:
                    pa[pi[k]].remove(l)

        return pa

    def estimate_total_effect(self, X, from_index, to_index):
        """Estimate total effect using causal model.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Original data, where n_samples is the number of samples
            and n_features is the number of features.
        from_index :
            Index of source variable to estimate total effect.
        to_index :
            Index of destination variable to estimate total effect.

        Returns
        -------
        total_effect : float
            **Because RESIT is a nonlinear algorithm, it cannot estimate the total effect and always returns a value of zero**
        """
        return 0

    def get_error_independence_p_values(self, X):
        """Calculate the p-value matrix of independence between error variables.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Original data, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        independence_p_values : array-like, shape (n_features, n_features)
            **RESIT always returns zero**
        """
        n_features = X.shape[1]
        p_values = np.zeros([n_features, n_features])
        return p_values



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""
import itertools
import numbers
import warnings

import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.utils import check_array

from .direct_lingam import DirectLiNGAM
from .hsic import hsic_test_gamma
from .utils import predict_adaptive_lasso, find_all_paths


class LongitudinalLiNGAM:
    """Implementation of Longitudinal LiNGAM algorithm [1]_

    References
    ----------
    .. [1] K. Kadowaki, S. Shimizu, and T. Washio. Estimation of causal structures
       in longitudinal data using non-Gaussianity. In Proc. 23rd IEEE International
       Workshop on Machine Learning for Signal Processing (MLSP2013), pp. 1--6, Southampton, United Kingdom, 2013.
    """

    def __init__(self, n_lags=1, measure="pwling", random_state=None):
        """Construct a model.

        Parameters
        ----------
        n_lags : int, optional (default=1)
            Number of lags.
        measure : {'pwling', 'kernel'}, default='pwling'
            Measure to evaluate independence : 'pwling' or 'kernel'.
        random_state : int, optional (default=None)
            ``random_state`` is the seed used by the random number generator.
        """
        self._n_lags = n_lags
        self._measure = measure
        self._random_state = random_state
        self._causal_orders = None
        self._adjacency_matrices = None

    def fit(self, X_list):
        """Fit the model to datasets.

        Parameters
        ----------
        X_list : list, shape [X, ...]
            Longitudinal multiple datasets for training, where ``X`` is an dataset.
            The shape of ``X`` is (n_samples, n_features),
            where ``n_samples`` is the number of samples and ``n_features`` is the number of features.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        # Check parameters
        if not isinstance(X_list, (list, np.ndarray)):
            raise ValueError("X_list must be a array-like.")

        if len(X_list) < 2:
            raise ValueError("X_list must be a list containing at least two items")

        self._T = len(X_list)
        self._n = check_array(X_list[0]).shape[0]
        self._p = check_array(X_list[0]).shape[1]
        X_t = []
        for X in X_list:
            X = check_array(X)
            if X.shape != (self._n, self._p):
                raise ValueError("X_list must be a list with the same shape")
            X_t.append(X.T)

        M_tau, N_t = self._compute_residuals(X_t)
        B_t, causal_orders = self._estimate_instantaneous_effects(N_t)
        B_tau = self._estimate_lagged_effects(B_t, M_tau)

        # output B(t,t), B(t,t-τ)
        self._adjacency_matrices = np.empty(
            (self._T, 1 + self._n_lags, self._p, self._p)
        )
        self._adjacency_matrices[:, :] = np.nan
        for t in range(1, self._T):
            self._adjacency_matrices[t, 0] = B_t[t]
            for l in range(self._n_lags):
                if t - l != 0:
                    self._adjacency_matrices[t, l + 1] = B_tau[t, l]

        self._residuals = np.zeros((self._T, self._n, self._p))
        for t in range(self._T):
            self._residuals[t] = N_t[t].T
        self._causal_orders = causal_orders
        return self

    def bootstrap(self, X_list, n_sampling, start_from_t=1):
        """Evaluate the statistical reliability of DAG based on the bootstrapping.

        Parameters
        ----------
        X_list : array-like, shape (X, ...)
            Longitudinal multiple datasets for training, where ``X`` is an dataset.
            The shape of ''X'' is (n_samples, n_features),
            where ``n_samples`` is the number of samples and ``n_features`` is the number of features.
        n_sampling : int
            Number of bootstrapping samples.

        Returns
        -------
        results : array-like, shape (BootstrapResult, ...)
            Returns the results of bootstrapping for multiple datasets.
        """
        # Check parameters
        if not isinstance(X_list, (list, np.ndarray)):
            raise ValueError("X_list must be a array-like.")

        if len(X_list) < 2:
            raise ValueError("X_list must be a list containing at least two items")

        self._T = len(X_list)
        self._n = check_array(X_list[0]).shape[0]
        self._p = check_array(X_list[0]).shape[1]
        X_t = []
        for X in X_list:
            X = check_array(X)
            if X.shape != (self._n, self._p):
                raise ValueError("X_list must be a list with the same shape")
            X_t.append(X)

        # Bootstrapping
        adjacency_matrices = np.zeros(
            (n_sampling, self._T, 1 + self._n_lags, self._p, self._p)
        )
        total_effects = np.zeros((n_sampling, self._T * self._p, self._T * self._p))
        for i in range(n_sampling):
            resampled_X_t = np.empty((self._T, self._n, self._p))
            indices = np.random.randint(0, self._n, size=(self._n,))
            for t in range(self._T):
                resampled_X_t[t] = X_t[t][indices, :]

            self.fit(resampled_X_t)
            adjacency_matrices[i] = self._adjacency_matrices

            # Calculate total effects
            for from_t in range(start_from_t, self._T):
                for c, from_ in enumerate(self._causal_orders[from_t]):
                    to_t = from_t
                    for to in self._causal_orders[from_t][c + 1 :]:
                        total_effects[
                            i, to_t * self._p + to, from_t * self._p + from_
                        ] = self.estimate_total_effect(resampled_X_t, from_t, from_, to_t, to)

                    for to_t in range(from_t + 1, self._T):
                        for to in self._causal_orders[to_t]:
                            total_effects[
                                i, to_t * self._p + to, from_t * self._p + from_
                            ] = self.estimate_total_effect(resampled_X_t, from_t, from_, to_t, to)

        return LongitudinalBootstrapResult(self._T, adjacency_matrices, total_effects)

    def estimate_total_effect(self, X_t, from_t, from_index, to_t, to_index):
        """Estimate total effect using causal model.

        Parameters
        ----------
        X_t : array-like, shape (n_samples, n_features)
            Original data, where n_samples is the number of samples
            and n_features is the number of features.
        from _t :
            The timepoint of source variable.
        from_index :
            Index of source variable to estimate total effect.
        to_t :
            The timepoint of destination variable.
        to_index :
            Index of destination variable to estimate total effect.

        Returns
        -------
        total_effect : float
            Estimated total effect.
        """
        # Check from/to causal order
        if to_t == from_t:
            from_order = self._causal_orders[to_t].index(from_index)
            to_order = self._causal_orders[from_t].index(to_index)
            if from_order > to_order:
                warnings.warn(
                    f"The estimated causal effect may be incorrect because "
                    f"the causal order of the destination variable (to_t={to_t}, to_index={to_index}) "
                    f"is earlier than the source variable (from_t={from_t}, from_index={from_index})."
                )
        elif to_t < from_t:
            warnings.warn(
                f"The estimated causal effect may be incorrect because "
                f"the causal order of the destination variable (to_t={to_t}) "
                f"is earlier than the source variable (from_t={from_t})."
            )

        # X + lagged X
        # n_features * (to + from + n_lags)
        X_joined = np.zeros((self._n, self._p * (2 + self._n_lags)))
        X_joined[:, 0 : self._p] = X_t[to_t]
        for tau in range(1 + self._n_lags):
            pos = self._p + self._p * tau
            X_joined[:, pos : pos + self._p] = X_t[from_t - tau]

        am = np.concatenate([*self._adjacency_matrices[from_t]], axis=1)

        # from_index + parents indices
        parents = np.where(np.abs(am[from_index]) > 0)[0]
        predictors = [from_index + self._p]
        predictors.extend(parents + self._p)

        # Estimate total effect
        coefs = predict_adaptive_lasso(X_joined, predictors, to_index)

        return coefs[0]

    def get_error_independence_p_values(self):
        """Calculate the p-value matrix of independence between error variables.

        Returns
        -------
        independence_p_values : array-like, shape (n_features, n_features)
            p-value matrix of independence between error variables.
        """
        E_list = np.empty((self._T, self._n, self._p))
        for t, resid in enumerate(self.residuals_):
            B_t = self._adjacency_matrices[t, 0]
            E_list[t] = np.dot(np.eye(B_t.shape[0]) - B_t, resid.T).T

        p_values_list = np.zeros([self._T, self._p, self._p])
        p_values_list[:, :, :] = np.nan
        for t in range(1, self._T):
            p_values = np.zeros([self._p, self._p])
            for i, j in itertools.combinations(range(self._p), 2):
                _, p_value = hsic_test_gamma(
                    np.reshape(E_list[t][:, i], [self._n, 1]),
                    np.reshape(E_list[t][:, j], [self._n, 1]),
                )
                p_values[i, j] = p_value
                p_values[j, i] = p_value

            p_values_list[t] = p_values

        return p_values_list

    def _compute_residuals(self, X_t):
        """Compute residuals N(t)"""
        M_tau = np.zeros((self._T, self._n_lags, self._p, self._p))
        N_t = np.zeros((self._T, self._p, self._n))
        N_t[:, :, :] = np.nan

        for t in range(1, self._T):
            # predictors
            X_predictors = np.zeros((self._n, self._p * (1 + self._n_lags)))
            for tau in range(self._n_lags):
                pos = self._p * tau
                X_predictors[:, pos : pos + self._p] = X_t[t - (tau + 1)].T

            # estimate M(t,t-τ) by regression
            X_target = X_t[t].T
            for i in range(self._p):
                reg = LinearRegression()
                reg.fit(X_predictors, X_target[:, i])
                for tau in range(self._n_lags):
                    pos = self._p * tau
                    M_tau[t, tau, i] = reg.coef_[pos : pos + self._p]

            # Compute N(t)
            N_t[t] = X_t[t]
            for tau in range(self._n_lags):
                N_t[t] = N_t[t] - np.dot(M_tau[t, tau], X_t[t - (tau + 1)])

        return M_tau, N_t

    def _estimate_instantaneous_effects(self, N_t):
        """Estimate instantaneous effects B(t,t) by applying LiNGAM"""
        causal_orders = [[np.nan] * self._p]
        B_t = np.zeros((self._T, self._p, self._p))
        for t in range(1, self._T):
            model = DirectLiNGAM(measure=self._measure)
            model.fit(N_t[t].T)
            causal_orders.append(model.causal_order_)
            B_t[t] = model.adjacency_matrix_
        return B_t, causal_orders

    def _estimate_lagged_effects(self, B_t, M_tau):
        """Estimate lagged effects B(t,t-τ)"""
        B_tau = np.zeros((self._T, self._n_lags, self._p, self._p))
        for t in range(self._T):
            for tau in range(self._n_lags):
                B_tau[t, tau] = np.dot(np.eye(self._p) - B_t[t], M_tau[t, tau])
        return B_tau

    @property
    def causal_orders_(self):
        """Estimated causal ordering.

        Returns
        -------
        causal_order_ : array-like, shape (causal_order, ...)
            The causal order of fitted models for B(t,t).
            The shape of causal_order is (n_features),
            where ``n_features`` is the number of features.
        """
        return self._causal_orders

    @property
    def adjacency_matrices_(self):
        """Estimated adjacency matrices.

        Returns
        -------
        adjacency_matrices_ : array-like, shape ((B(t,t), B(t,t-1), ..., B(t,t-τ)), ...)
            The list of adjacency matrix B(t,t) and B(t,t-τ) for longitudinal datasets.
            The shape of B(t,t) and B(t,t-τ) is (n_features, n_features), where
            ``n_features`` is the number of features.
            **If the previous data required for the calculation are not available,
            such as B(t,t) or B(t,t-τ) at t=0, all elements of the matrix are nan**.
        """
        return self._adjacency_matrices

    @property
    def residuals_(self):
        """Residuals of regression.

        Returns
        -------
        residuals_ : list, shape [E, ...]
            Residuals of regression, where ``E`` is an dataset.
            The shape of ``E`` is (n_samples, n_features),
            where ``n_samples`` is the number of samples and ``n_features`` is the number of features.

        """
        return self._residuals


class LongitudinalBootstrapResult(object):
    """The result of bootstrapping for LongitudinalLiNGAM."""

    def __init__(self, n_timepoints, adjacency_matrices, total_effects):
        """Construct a BootstrapResult.

        Parameters
        ----------
        adjacency_matrices : array-like, shape (n_sampling)
            The adjacency matrix list by bootstrapping.
        total_effects : array-like, shape (n_sampling)
            The total effects list by bootstrapping.
        """
        self._n_timepoints = n_timepoints
        self._adjacency_matrices = adjacency_matrices
        self._total_effects = total_effects

    @property
    def adjacency_matrices_(self):
        """The adjacency matrix list by bootstrapping.

        Returns
        -------
        adjacency_matrices_ : array-like, shape (n_sampling)
            The adjacency matrix list, where ``n_sampling`` is
            the number of bootstrap sampling.
        """
        return self._adjacency_matrices

    @property
    def total_effects_(self):
        """The total effect list by bootstrapping.

        Returns
        -------
        total_effects_ : array-like, shape (n_sampling)
            The total effect list, where ``n_sampling`` is
            the number of bootstrap sampling.
        """
        return self._total_effects

    def get_causal_direction_counts(
        self,
        n_directions=None,
        min_causal_effect=None,
        split_by_causal_effect_sign=False,
    ):
        """Get causal direction count as a result of bootstrapping.

        Parameters
        ----------
        n_directions : int, optional (default=None)
            If int, then The top ``n_directions`` items are included in the result
        min_causal_effect : float, optional (default=None)
            Threshold for detecting causal direction.
            If float, then causal directions with absolute values of causal
            effects less than ``min_causal_effect`` are excluded.
        split_by_causal_effect_sign : boolean, optional (default=False)
            If True, then causal directions are split depending on the sign of the causal effect.

        Returns
        -------
        causal_direction_counts : dict
            List of causal directions sorted by count in descending order.
            The dictionary has the following format::

            {'from': [n_directions], 'to': [n_directions], 'count': [n_directions]}

            where ``n_directions`` is the number of causal directions.
        """
        # Check parameters
        if isinstance(n_directions, (numbers.Integral, np.integer)):
            if not 0 < n_directions:
                raise ValueError("n_directions must be an integer greater than 0")
        elif n_directions is None:
            pass
        else:
            raise ValueError("n_directions must be an integer greater than 0")

        if min_causal_effect is None:
            min_causal_effect = 0.0
        else:
            if not 0.0 < min_causal_effect:
                raise ValueError("min_causal_effect must be an value greater than 0.")

        # Count causal directions
        cdc_list = []
        for t in range(self._n_timepoints):

            directions = []
            for m in self._adjacency_matrices:
                am = np.concatenate([*m[t]], axis=1)
                direction = np.array(np.where(np.abs(am) > min_causal_effect))
                if split_by_causal_effect_sign:
                    signs = (
                        np.array([np.sign(am[i][j]) for i, j in direction.T])
                        .astype("int64")
                        .T
                    )
                    direction = np.vstack([direction, signs])
                directions.append(direction.T)
            directions = np.concatenate(directions)

            if len(directions) == 0:
                cdc = {"from": [], "to": [], "count": []}
                if split_by_causal_effect_sign:
                    cdc["sign"] = []
                cdc_list.append(cdc)
                continue

            directions, counts = np.unique(directions, axis=0, return_counts=True)
            sort_order = np.argsort(-counts)
            sort_order = (
                sort_order[:n_directions] if n_directions is not None else sort_order
            )
            counts = counts[sort_order]
            directions = directions[sort_order]

            cdc = {
                "from": directions[:, 1].tolist(),
                "to": directions[:, 0].tolist(),
                "count": counts.tolist(),
            }
            if split_by_causal_effect_sign:
                cdc["sign"] = directions[:, 2].tolist()

            cdc_list.append(cdc)

        return cdc_list

    def get_directed_acyclic_graph_counts(
        self, n_dags=None, min_causal_effect=None, split_by_causal_effect_sign=False
    ):
        """Get DAGs count as a result of bootstrapping.

        Parameters
        ----------
        n_dags : int, optional (default=None)
            If int, then The top ``n_dags`` items are included in the result
        min_causal_effect : float, optional (default=None)
            Threshold for detecting causal direction.
            If float, then causal directions with absolute values of causal effects less than
            ``min_causal_effect`` are excluded.
        split_by_causal_effect_sign : boolean, optional (default=False)
            If True, then causal directions are split depending on the sign of the causal effect.

        Returns
        -------
        directed_acyclic_graph_counts : dict
            List of directed acyclic graphs sorted by count in descending order.
            The dictionary has the following format::

            {'dag': [n_dags], 'count': [n_dags]}.

            where ``n_dags`` is the number of directed acyclic graphs.
        """
        # Check parameters
        if isinstance(n_dags, (numbers.Integral, np.integer)):
            if not 0 < n_dags:
                raise ValueError("n_dags must be an integer greater than 0")
        elif n_dags is None:
            pass
        else:
            raise ValueError("n_dags must be an integer greater than 0")

        if min_causal_effect is None:
            min_causal_effect = 0.0
        else:
            if not 0.0 < min_causal_effect:
                raise ValueError("min_causal_effect must be an value greater than 0.")

        # Count directed acyclic graphs
        dagc_list = []
        for t in range(self._n_timepoints):

            dags = []
            for m in self._adjacency_matrices:
                am = np.concatenate([*m[t]], axis=1)

                dag = np.abs(am) > min_causal_effect
                if split_by_causal_effect_sign:
                    direction = np.array(np.where(dag))
                    signs = np.zeros_like(dag).astype("int64")
                    for i, j in direction.T:
                        signs[i][j] = np.sign(am[i][j]).astype("int64")
                    dag = signs
                dags.append(dag)

            dags, counts = np.unique(dags, axis=0, return_counts=True)
            sort_order = np.argsort(-counts)
            sort_order = sort_order[:n_dags] if n_dags is not None else sort_order
            counts = counts[sort_order]
            dags = dags[sort_order]

            if split_by_causal_effect_sign:
                dags = [
                    {
                        "from": np.where(dag)[1].tolist(),
                        "to": np.where(dag)[0].tolist(),
                        "sign": [dag[i][j] for i, j in np.array(np.where(dag)).T],
                    }
                    for dag in dags
                ]
            else:
                dags = [
                    {"from": np.where(dag)[1].tolist(), "to": np.where(dag)[0].tolist()}
                    for dag in dags
                ]

            dagc_list.append({"dag": dags, "count": counts.tolist()})

        return dagc_list

    def get_probabilities(self, min_causal_effect=None):
        """Get bootstrap probability.

        Parameters
        ----------
        min_causal_effect : float, optional (default=None)
            Threshold for detecting causal direction.
            If float, then causal directions with absolute values of causal effects less than
            ``min_causal_effect`` are excluded.

        Returns
        -------
        probabilities : array-like
            List of bootstrap probability matrix.
        """
        # check parameters
        if min_causal_effect is None:
            min_causal_effect = 0.0
        else:
            if not 0.0 < min_causal_effect:
                raise ValueError("min_causal_effect must be an value greater than 0.")

        prob = np.zeros(self._adjacency_matrices[0].shape)
        for adj_mat in self._adjacency_matrices:
            prob += np.where(np.abs(adj_mat) > min_causal_effect, 1, 0)
        prob = prob / len(self._adjacency_matrices)

        return prob

    def get_total_causal_effects(self, min_causal_effect=None):
        """Get total effects list.

        Parameters
        ----------
        min_causal_effect : float, optional (default=None)
            Threshold for detecting causal direction.
            If float, then causal directions with absolute values of causal effects less than
            ``min_causal_effect`` are excluded.

        Returns
        -------
        total_causal_effects : dict
            List of bootstrap total causal effect sorted by probability in descending order.
            The dictionary has the following format::

            {'from': [n_directions], 'to': [n_directions], 'effect': [n_directions], 'probability': [n_directions]}

            where ``n_directions`` is the number of causal directions.
        """
        # Check parameters
        if min_causal_effect is None:
            min_causal_effect = 0.0
        else:
            if not 0.0 < min_causal_effect:
                raise ValueError("min_causal_effect must be an value greater than 0.")

        # probability
        probs = np.sum(
            np.where(np.abs(self._total_effects) > min_causal_effect, 1, 0),
            axis=0,
            keepdims=True,
        )[0]
        probs = probs / len(self._total_effects)

        # causal directions
        dirs = np.array(np.where(np.abs(probs) > 0))
        probs = probs[dirs[0], dirs[1]]

        # calculate median effect without zero
        effects = np.zeros(dirs.shape[1])
        for i, (to, from_) in enumerate(dirs.T):
            idx = np.where(np.abs(self._total_effects[:, to, from_]) > 0)
            effects[i] = np.median(self._total_effects[:, to, from_][idx])

        # sort by effect value
        order = np.argsort(-probs)
        dirs = dirs.T[order]
        effects = effects[order]
        probs = probs[order]

        ce = {
            "from": dirs[:, 1].tolist(),
            "to": dirs[:, 0].tolist(),
            "effect": effects.tolist(),
            "probability": probs.tolist(),
        }

        return ce

    def get_paths(self, from_index, to_index, from_t, to_t, min_causal_effect=None):
        """Get all paths from the start variable to the end variable and their bootstrap probabilities.

        Parameters
        ----------
        from_index : int
            Index of the variable at the start of the path.
        to_index : int
            Index of the variable at the end of the path.
        from_t : int
            The starting timepoint of the path.
        to_t : int
            The end timepoint of the path.
        min_causal_effect : float, optional (default=None)
            Threshold for detecting causal direction.
            Causal directions with absolute values of causal effects less than ``min_causal_effect`` are excluded.

        Returns
        -------
        paths : dict
            List of path and bootstrap probability.
            The dictionary has the following format::

            {'path': [n_paths], 'effect': [n_paths], 'probability': [n_paths]}

            where ``n_paths`` is the number of paths.
        """
        # check parameters
        if min_causal_effect is None:
            min_causal_effect = 0.0
        else:
            if not 0.0 < min_causal_effect:
                raise ValueError("min_causal_effect must be an value greater than 0.")
        if to_t < from_t:
            raise ValueError("to_t should be greater than or equal to from_t.")
        if to_t == from_t:
            if to_index == from_index:
                raise ValueError("The same variable is specified for from and to.")

        # Find all paths from from_index to to_index
        paths_list = []
        effects_list = []
        for am in self._adjacency_matrices:
            n_timepoints = to_t - from_t + 1
            n_features = am.shape[2]
            expansion_m_size = n_features * n_timepoints
            expansion_m = np.zeros((expansion_m_size, expansion_m_size))
            for i in range(n_timepoints):
                for j in range(n_timepoints):
                    row = n_features * i
                    col = n_features * j
                    if col > row:
                        continue
                    expansion_m[row : row + n_features, col : col + n_features] = am[
                        i + from_t, i - j
                    ]

            paths, effects = find_all_paths(
                expansion_m,
                int(from_index),
                int(n_features * (to_t - from_t) + to_index),
                min_causal_effect,
            )

            # Convert path to string to make them easier to handle.
            paths_list.extend(["_".join(map(str, p)) for p in paths])
            effects_list.extend(effects)

        paths_list = np.array(paths_list)
        effects_list = np.array(effects_list)

        # Count paths
        paths_str, counts = np.unique(paths_list, axis=0, return_counts=True)

        # Sort by count
        order = np.argsort(-counts)
        probs = counts[order] / len(self._adjacency_matrices)
        paths_str = paths_str[order]

        # Calculate median of causal effect for each path
        effects = [
            np.median(effects_list[np.where(paths_list == p)]) for p in paths_str
        ]

        result = {
            "path": [[int(i) for i in p.split("_")] for p in paths_str],
            "effect": effects,
            "probability": probs.tolist(),
        }
        return result



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""
import graphviz
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from matplotlib import colors as mcolors
from matplotlib.colors import is_color_like
from sklearn import linear_model
from sklearn.linear_model import LassoLarsIC, LinearRegression
from sklearn.preprocessing import StandardScaler
from sklearn.utils import check_array, check_scalar

from ._rcd import extract_ancestors
from ._f_correlation import f_correlation
from ._visualize_nonlinear_causal_effect import visualize_nonlinear_causal_effect

__all__ = [
    "print_causal_directions",
    "print_dagc",
    "make_prior_knowledge",
    "remove_effect",
    "make_dot",
    "make_dot_highlight",
    "predict_adaptive_lasso",
    "get_sink_variables",
    "get_exo_variables",
    "find_all_paths",
    "likelihood_i",
    "log_p_super_gaussian",
    "variance_i",
    "extract_ancestors",
    "f_correlation",
    "visualize_nonlinear_causal_effect",
]


def print_causal_directions(cdc, n_sampling, labels=None):
    """Print causal directions of bootstrap result to stdout.

    Parameters
    ----------
    cdc : dict
        List of causal directions sorted by count in descending order.
        This can be set the value returned by ``BootstrapResult.get_causal_direction_counts()`` method.
    n_sampling : int
        Number of bootstrapping samples.
    labels : array-like, optional (default=None)
        List of feature lables.
        If set labels, the output feature name will be the specified label.
    """
    for i, (fr, to, co) in enumerate(zip(cdc["from"], cdc["to"], cdc["count"])):
        sign = "" if "sign" not in cdc else "(b>0)" if cdc["sign"][i] > 0 else "(b<0)"
        if labels:
            print(f"{labels[to]} <--- {labels[fr]} {sign} ({100*co/n_sampling:.1f}%)")
        else:
            print(f"x{to} <--- x{fr} {sign} ({100*co/n_sampling:.1f}%)")


def print_dagc(dagc, n_sampling, labels=None):
    """Print DAGs of bootstrap result to stdout.

    Parameters
    ----------
    dagc : dict
        List of directed acyclic graphs sorted by count in descending order.
        This can be set the value returned by ``BootstrapResult.get_directed_acyclic_graph_counts()`` method.
    n_sampling : int
        Number of bootstrapping samples.
    labels : array-like, optional (default=None)
        List of feature lables.
        If set labels, the output feature name will be the specified label.
    """
    for i, (dag, co) in enumerate(zip(dagc["dag"], dagc["count"])):
        print(f"DAG[{i}]: {100*co/n_sampling:.1f}%")
        for j, (fr, to) in enumerate(zip(dag["from"], dag["to"])):
            sign = (
                "" if "sign" not in dag else "(b>0)" if dag["sign"][j] > 0 else "(b<0)"
            )
            if labels:
                print("\t" + f"{labels[to]} <--- {labels[fr]} {sign}")
            else:
                print("\t" + f"x{to} <--- x{fr} {sign}")


def make_prior_knowledge(
    n_variables,
    exogenous_variables=None,
    sink_variables=None,
    paths=None,
    no_paths=None,
):
    """Make matrix of prior knowledge.

    Parameters
    ----------
    n_variables : int
        Number of variables.
    exogenous_variables : array-like, shape (index, ...), optional (default=None)
        List of exogenous variables(index).
        Prior knowledge is created with the specified variables as exogenous variables.
    sink_variables : array-like, shape (index, ...), optional (default=None)
        List of sink variables(index).
        Prior knowledge is created with the specified variables as sink variables.
    paths : array-like, shape ((index, index), ...), optional (default=None)
        List of variables(index) pairs with directed path.
        If ``(i, j)``, prior knowledge is created that xi has a directed path to xj.
    no_paths : array-like, shape ((index, index), ...), optional (default=None)
        List of variables(index) pairs without directed path.
        If ``(i, j)``, prior knowledge is created that xi does not have a directed path to xj.

    Returns
    -------
    prior_knowledge : array-like, shape (n_variables, n_variables)
        Return matrix of prior knowledge used for causal discovery.
    """
    prior_knowledge = np.full((n_variables, n_variables), -1)
    if no_paths:
        for no_path in no_paths:
            prior_knowledge[no_path[1], no_path[0]] = 0
    if paths:
        for path in paths:
            prior_knowledge[path[1], path[0]] = 1
    if sink_variables:
        for var in sink_variables:
            prior_knowledge[:, var] = 0
    if exogenous_variables:
        for var in exogenous_variables:
            prior_knowledge[var, :] = 0
    np.fill_diagonal(prior_knowledge, -1)
    return prior_knowledge


def get_sink_variables(adjacency_matrix):
    """The sink variables(index) in the adjacency matrix.

    Parameters
    ----------
    adjacency_matrix : array-like, shape (n_variables, n_variables)
        Adjacency matrix, where n_variables is the number of variables.

    Returns
    -------
    sink_variables : array-like
        List of sink variables(index).
    """
    am = adjacency_matrix.copy()
    am = np.abs(am)
    np.fill_diagonal(am, 0)
    sink_vars = [i for i in range(am.shape[1]) if am[:, i].sum() == 0]
    return sink_vars


def get_exo_variables(adjacency_matrix):
    """The exogenous variables(index) in the adjacency matrix.

    Parameters
    ----------
    adjacency_matrix : array-like, shape (n_variables, n_variables)
        Adjacency matrix, where n_variables is the number of variables.

    Returns
    -------
    exogenous_variables : array-like
        List of exogenous variables(index).
    """
    am = adjacency_matrix.copy()
    am = np.abs(am)
    np.fill_diagonal(am, 0)
    exo_vars = [i for i in range(am.shape[1]) if am[i, :].sum() == 0]
    return exo_vars


def remove_effect(X, remove_features):
    """Create a dataset that removes the effects of features by linear regression.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Data, where ``n_samples`` is the number of samples
        and ``n_features`` is the number of features.
    remove_features : array-like
        List of features(index) to remove effects.

    Returns
    -------
    X : array-like, shape (n_samples, n_features)
        Data after removing effects of ``remove_features``.
    """
    X = np.copy(check_array(X))
    features_ = [i for i in np.arange(X.shape[1]) if i not in remove_features]
    for feature in features_:
        reg = linear_model.LinearRegression()
        reg.fit(X[:, remove_features], X[:, feature])
        X[:, feature] = X[:, feature] - reg.predict(X[:, remove_features])
    return X


def make_dot(
    adjacency_matrix,
    labels=None,
    lower_limit=0.01,
    prediction_feature_indices=None,
    prediction_target_label="Y(pred)",
    prediction_line_color="red",
    prediction_coefs=None,
    prediction_feature_importance=None,
    path=None,
    path_color=None,
    detect_cycle=False,
    ignore_shape=False,
):
    """Directed graph source code in the DOT language with specified adjacency matrix.

    Parameters
    ----------
    adjacency_matrix : array-like with shape (n_features, n_features)
        Adjacency matrix to make graph, where ``n_features`` is the number of features.
    labels : array-like, optional (default=None)
        Label to use for graph features.
    lower_limit : float, optional (default=0.01)
        Threshold for drawing direction.
        If float, then directions with absolute values of coefficients less than ``lower_limit`` are excluded.
    prediction_feature_indices : array-like, optional (default=None)
        Indices to use as prediction features.
    prediction_target_label : string, optional (default='Y(pred)'))
        Label to use for target variable of prediction.
    prediction_line_color : string, optional (default='red')
        Line color to use for prediction's graph.
    prediction_coefs : array-like, optional (default=None)
        Coefficients to use for prediction's graph.
    prediction_feature_importance : array-like, optional (default=None)
        Feature importance to use for prediction's graph.
    path : tuple, optional (default=None)
        Path to highlight. Tuple of start index and end index.
    path_color : string, optional (default=None)
        Colors to highlight a path.
    detect_cycle : boolean, optional (default=False)
        Highlight simple cycles.
    ignore_shape : boolean, optional (default=False)
        Ignore checking the shape of adjaceny_matrix or not.

    Returns
    -------
    graph : graphviz.Digraph
        Directed graph source code in the DOT language.
        If order is unknown, draw a double-headed arrow.
    """
    # Check parameters
    B = check_array(np.nan_to_num(adjacency_matrix))
    if not ignore_shape and B.shape[0] != B.shape[1]:
        raise ValueError("'adjacency_matrix' is not square matrix.")
    if labels is not None:
        if B.shape[1] != len(labels):
            raise ValueError(
                "Length of 'labels' does not match length of 'adjacency_matrix'"
            )
    if prediction_feature_indices is not None:
        if prediction_coefs is not None and (
            len(prediction_feature_indices) != len(prediction_coefs)
        ):
            raise ValueError(
                "Length of 'prediction_coefs' does not match length of 'prediction_feature_indices'"
            )
        if prediction_feature_importance is not None and (
            len(prediction_feature_indices) != len(prediction_feature_importance)
        ):
            raise ValueError(
                "Length of 'prediction_feature_importance' does not match length of 'prediction_feature_indices'"
            )

    if path is not None:
        if not isinstance(path, tuple) or len(path) < 2:
            raise TypeError("'path' should be a tuple of node indices.")
        if path_color is not None and not is_color_like(path_color):
            raise ValueError("'path_color' should be an color name.")

        if path_color is None:
            path_color = "black"
        path_start = check_scalar(
            path[0], "path_start", int, min_val=0, max_val=B.shape[0]
        )
        path_end = check_scalar(path[1], "path_end", int, min_val=0, max_val=B.shape[0])

    if detect_cycle is not None:
        if not isinstance(detect_cycle, bool):
            raise TypeError("'detect_cycle' should be boolean.")

    idx = np.abs(B) > lower_limit

    nx_graph = None
    if path is not None or detect_cycle is True:
        nx_graph = nx.from_numpy_array(idx.astype(int).T, create_using=nx.DiGraph)

    # search path
    path_edges = None
    path_nodes = None
    if path is not None:
        path_edges = set()
        for path_ in nx.all_simple_paths(nx_graph, path[0], path[1]):
            path_edges |= set(zip(path_, path_[1:]))

        path_nodes = np.unique([list(edge) for edge in path_edges])
        if labels is None:
            path_nodes = [f"x{node}" for node in path_nodes]
        else:
            path_nodes = [labels[node] for node in path_nodes]

    # simple cycles
    cycle_edges = []
    if detect_cycle is True:
        for cycle in nx.simple_cycles(nx_graph):
            es = list(zip(cycle, np.roll(cycle, -1)))
            cycle_edges.append(es)
        cycle_edges = sum(cycle_edges, start=[])

    d = graphviz.Digraph(engine="dot")

    # nodes
    names = labels if labels else [f"x{i}" for i in range(len(B))]
    for name in names:
        kwargs = {}
        if path is not None:
            if name in path_nodes:
                kwargs["fontcolor"] = path_color
                kwargs["color"] = path_color
            if name == names[path_start] or name == names[path_end]:
                kwargs["peripheries"] = "2"
        d.node(name, **kwargs)

    # edges
    dirs = np.where(idx)
    for to, from_, coef in zip(dirs[0], dirs[1], B[idx]):
        kwargs = {}
        if path is not None and (from_, to) in path_edges:
            kwargs["penwidth"] = "2"
            kwargs["fontcolor"] = path_color
            kwargs["color"] = path_color
        if detect_cycle is True and (from_, to) in cycle_edges:
            kwargs["style"] = "dashed"
        d.edge(names[from_], names[to], label=f"{coef:.2f}", **kwargs)

    # integrate of prediction model
    if prediction_feature_indices is not None:
        d.node(
            prediction_target_label,
            color=prediction_line_color,
            fontcolor=prediction_line_color,
        )

        if prediction_coefs is not None:
            for from_, coef in zip(prediction_feature_indices, prediction_coefs):
                if np.abs(coef) > lower_limit:
                    d.edge(
                        names[from_],
                        prediction_target_label,
                        label=f"{coef:.2f}",
                        color=prediction_line_color,
                        fontcolor=prediction_line_color,
                        style="dashed",
                    )

        elif prediction_feature_importance is not None:
            for from_, imp in zip(
                prediction_feature_indices, prediction_feature_importance
            ):
                d.edge(
                    names[from_],
                    prediction_target_label,
                    label=f"({imp})",
                    color=prediction_line_color,
                    fontcolor=prediction_line_color,
                    style="dashed",
                )

        else:
            for from_ in prediction_feature_indices:
                d.edge(
                    names[from_],
                    prediction_target_label,
                    color=prediction_line_color,
                    style="dashed",
                )

    # If the value is nan, draw a double-headed arrow
    unk_order = np.where(np.isnan(np.tril(adjacency_matrix)))
    unk_order_set = set([val for item in unk_order for val in item])
    with d.subgraph() as s:
        s.attr(rank="same")
        for node in unk_order_set:
            s.node(names[node])
    for to, from_ in zip(unk_order[0], unk_order[1]):
        d.edge(names[from_], names[to], dir="both")

    return d


def make_dot_highlight(
    adjacency_matrix,
    node_index,
    labels=None,
    max_dsc=None,
    max_anc=None,
    lower_limit=None,
    cmap=None,
    draw_others=True,
    detect_cycle=False,
    path=None,
    path_color=None,
    vmargin=2,
    hmargin=2,
):
    """Hierachical make_dot.

    Parameters
    ----------
    adjacency_matrix : array-like with shape (n_features, n_features)
        Adjacency matrix to make graph, where ``n_features`` is the number of features.
    node_index : int
        Index of target node.
    labels : array-like, optional (default=None)
        Label to use for graph features.
    max_dsc : int, optional (default=None)
        Number of hierarchies on the descendant side to display.
    max_asc : int, optional (default=None)
        Number of hierarchies on the ancestral side to display.
    lower_limit : float, optional (default=0.01)
        Threshold for drawing direction.
        If float, then directions with absolute values of coefficients less than ``lower_limit`` are excluded.
    cmap : colormap, optional (default=None)
        Gradient color to represent hierarchy.
    draw_others : boolean, optional (default=True)
        Whether to draw others cluster.
    detect_cycle : boolean, optional (default=False)
        Highlight simple cycles.
    path : tuple, optional (default=None)
        Path to highlight. Tuple of start index and end index.
    path_color : string, optional (default=None)
        Colors to highlight a path.
    vmargin : float, optional (default=2)
        Vertical margin between nodes.
    hmargin : float, optional (default=2)
        Horizontal margin between nodes.

    Returns
    -------
    graph : graphviz.Digraph
        Directed graph source code in the DOT language.
        If order is unknown, draw a double-headed arrow.
    """

    # margins for clusters
    cls_margin_l = 1.5
    cls_margin_r = 0.6
    cls_margin_t = 0.6
    cls_margin_b = 0.4
    cls_space_target_other = 0.2

    # check arguments
    adj = check_array(
        adjacency_matrix,
        ensure_2d=True,
        ensure_min_samples=2,
        ensure_min_features=2,
        copy=True,
    )
    if adj.shape[0] != adj.shape[1]:
        raise ValueError

    node_index = check_scalar(
        node_index,
        "node_index",
        int,
        min_val=0,
        max_val=adj.shape[0],
        include_boundaries="left",
    )

    if labels is None:
        node_names = [f"x{i}" for i in range(adj.shape[1])]
    else:
        node_names = check_array(labels, ensure_2d=False, dtype=str)
        if len(node_names) != adj.shape[1]:
            raise ValueError

    if max_dsc is not None:
        max_dsc = check_scalar(max_dsc, "max_dsc", int, min_val=0)

    if max_anc is not None:
        max_anc = check_scalar(max_anc, "max_anc", int, min_val=0)

    if lower_limit is not None:
        lower_limit = check_scalar(lower_limit, "lower_limit", (int, float))

    draw_others = check_scalar(draw_others, "draw_others", bool)
    detect_cycle = check_scalar(detect_cycle, "detect_cycle", bool)

    path_start = None
    path_end = None
    if path is not None:
        if not isinstance(path, tuple) or len(path) < 2:
            raise TypeError("'path' should be a tuple of node indices.")
        if path_color is not None and not is_color_like(path_color):
            raise ValueError("'path_color' should be an color name.")
        if path_color is None:
            path_color = "black"
        path_start = check_scalar(
            path[0], "start", int, min_val=0, max_val=adj.shape[0]
        )
        path_end = check_scalar(path[1], "end", int, min_val=0, max_val=adj.shape[0])

    vmargin = check_scalar(vmargin, "vmargin", (float, int), min_val=1)
    hmargin = check_scalar(hmargin, "hmargin", (float, int), min_val=1)

    # apply lower_limit
    if lower_limit is not None:
        adj[abs(adj) < lower_limit] = 0

    # analyze network
    G = nx.from_numpy_array(~np.isclose(adj.T, 0), create_using=nx.DiGraph)

    dsc_path_length = nx.single_source_shortest_path_length(
        G, node_index, cutoff=max_dsc
    )
    del dsc_path_length[node_index]

    anc_path_length = nx.single_source_shortest_path_length(
        G.reverse(), node_index, cutoff=max_anc
    )
    del anc_path_length[node_index]

    # node indices
    dsc_indices = set(dsc_path_length.keys())
    anc_indices = set(anc_path_length.keys())
    isolate_indices = set(nx.isolates(G))
    other_indices = (
        set(np.arange(adj.shape[0]))
        - dsc_indices
        - anc_indices
        - isolate_indices
        - set([node_index])
    )

    # clusters (distance -> list of nodes)
    clusters = {}

    clusters[0] = [node_index]

    clusters[None] = sorted(other_indices) + sorted(isolate_indices)

    for node in dsc_indices:
        d = dsc_path_length[node]
        if d not in clusters.keys():
            clusters[d] = []
        clusters[d].append(node)

    for node in anc_indices:
        d = -anc_path_length[node]
        if d not in clusters.keys():
            clusters[d] = []
        clusters[d].append(node)

    # search path
    path_edges = None
    path_nodes = None
    if path is not None:
        path_edges = set()
        for path_ in nx.all_simple_paths(G, path_start, path_end):
            path_edges |= set(zip(path_, path_[1:]))

        path_nodes = np.unique([list(edge) for edge in path_edges])
        path_nodes = [node_names[node] for node in path_nodes]

    # create graphviz graph
    graph_attr = {"rankdir": "TB", "splines": "true"}
    graph = graphviz.Digraph("graph", engine="neato", graph_attr=graph_attr)

    # colors of clusters
    ds = sorted([d for d in clusters.keys() if d is not None])
    if cmap is None:
        cluster_colors = {d: "black" for d in ds}
    else:
        cmap = plt.get_cmap(cmap)
        cs = np.linspace(0, 1, len(ds) + 1)
        cluster_colors = dict(zip(ds, cmap(cs)))
    cluster_colors[None] = cluster_colors[0]

    # distance -> y_position
    pos_y_map = {0: 0, None: 0}

    # distance from target to descendants -> pos_y
    dsc_dists = sorted([d for d in clusters.keys() if d is not None and 0 < d])
    for i, dist in enumerate(dsc_dists):
        pos_y_map[dist] = -(i + 1) * vmargin

    # distance from target to ascendants -> pos_y
    anc_dists = sorted([d for d in clusters.keys() if d is not None and d < 0], key=abs)
    for i, dist in enumerate(anc_dists):
        pos_y_map[dist] = (i + 1) * vmargin

    get_cluster_name = lambda dist: f"cluster{dist}"
    get_node_id = lambda cl, node: f"{cls_name}-{node}"

    node_ids = {node: [] for node in np.arange(adj.shape[0])}

    # add subgraphs and nodes
    for dist, nodes in clusters.items():
        # skip drawing the cluster
        if dist is None and draw_others is False:
            continue

        if len(nodes) == 0:
            continue

        # attributes for the cluster and nodes
        if dist == 0:
            cls_label = "target"
            offset_x = 0
        elif dist is None:
            cls_label = "others"
            offset_x = cls_margin_r + cls_margin_l + cls_space_target_other
            nodes = sorted(
                nodes, key=lambda x: (x in isolate_indices, x in other_indices, x)
            )
        else:
            cls_label = (
                f"ancestor {abs(dist)}" if dist < 0 else f"descendant {abs(dist)}"
            )
            offset_x = 0
            nodes = sorted(nodes)
        pos_y = pos_y_map[dist]
        color = mcolors.to_hex(cluster_colors[dist], keep_alpha=True)
        cls_name = get_cluster_name(dist)

        graph_attr = {
            "style": "dashed",
            "label": cls_label,
            "labeljust": "l",
            "color": color,
        }
        with graph.subgraph(name=cls_name, graph_attr=graph_attr) as c:
            # add nodes
            for i, node in enumerate(nodes):
                node_id = get_node_id(cls_name, node)
                pos_x = hmargin * i + offset_x
                kwargs = {}
                if path is not None:
                    if node in path_nodes:
                        kwargs["fontcolor"] = path_color
                        kwargs["color"] = path_color
                    if node == path_start or node == path_end:
                        kwargs["peripheries"] = "2"
                c.node(
                    node_id,
                    node_names[node],
                    pos=f"{pos_x},{pos_y}!",
                    color=color,
                    **kwargs,
                )
                node_ids[node].append((node_id, dist))

            # dummy at the upper left
            x = -cls_margin_l + offset_x
            y = pos_y + cls_margin_t
            c.node(
                f"{cls_name}-dummy_upper",
                "",
                pos=f"{x},{y}!",
                style="invis",
                fixedsize="true",
                width="0",
                height="0",
            )

            # dummy at the lower right
            x = (len(nodes) - 1) * hmargin + cls_margin_r + offset_x
            y = pos_y - cls_margin_b
            c.node(
                f"{cls_name}-dummy_lower",
                "",
                pos=f"{x},{y}!",
                style="invis",
                fixedsize="true",
                width="0",
                height="0",
            )

    # cycles
    cycle_edges = []
    if detect_cycle is True:
        for cycle in nx.simple_cycles(G):
            es = list(zip(cycle, np.roll(cycle, -1)))
            cycle_edges.append(es)
        cycle_edges = sum(cycle_edges, start=[])

    # indices of clusters
    other_nodes = isolate_indices | other_indices
    valid_nodes = other_nodes | dsc_indices | anc_indices | set([node_index])

    # add edges
    edges = np.argwhere(~np.isclose(adj, 0))
    for h, t in edges:
        if draw_others is False and (t in other_nodes or h in other_nodes):
            continue

        if t not in valid_nodes or h not in valid_nodes:
            continue

        for tail_id, dist_t in node_ids[t]:
            for head_id, dist_h in node_ids[h]:
                color = cluster_colors[dist_h]
                color = mcolors.to_hex(color, keep_alpha=True)

                kwargs = {"color": color}
                if path is not None and (t, h) in path_edges:
                    kwargs["penwidth"] = "2"
                    kwargs["fontcolor"] = path_color
                    kwargs["color"] = path_color
                if detect_cycle is True and (t, h) in cycle_edges:
                    kwargs["style"] = "dashed"

                graph.edge(tail_id, head_id, label=f"{adj[h, t]:.2f}", **kwargs)

    return graph


def predict_adaptive_lasso(X, predictors, target, gamma=1.0):
    """Predict with Adaptive Lasso.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Training data, where n_samples is the number of samples
        and n_features is the number of features.
    predictors : array-like, shape (n_predictors)
        Indices of predictor variable.
    target : int
        Index of target variable.

    Returns
    -------
    coef : array-like, shape (n_features)
        Coefficients of predictor variable.
    """
    # Standardize X
    scaler = StandardScaler()
    X_std = scaler.fit_transform(X)

    # Pruning with Adaptive Lasso
    lr = LinearRegression()
    lr.fit(X_std[:, predictors], X_std[:, target])
    weight = np.power(np.abs(lr.coef_), gamma)
    reg = LassoLarsIC(criterion="bic")
    reg.fit(X_std[:, predictors] * weight, X_std[:, target])
    pruned_idx = np.abs(reg.coef_ * weight) > 0.0

    # Calculate coefficients of the original scale
    coef = np.zeros(reg.coef_.shape)
    if pruned_idx.sum() > 0:
        lr = LinearRegression()
        pred = np.array(predictors)
        lr.fit(X[:, pred[pruned_idx]], X[:, target])
        coef[pruned_idx] = lr.coef_

    return coef


def find_all_paths(dag, from_index, to_index, min_causal_effect=0.0):
    """Find all paths from point to point in DAG.

    Parameters
    ----------
    dag : array-like, shape (n_features, n_features)
        The adjacency matrix to fine all paths, where n_features is the number of features.
    from_index : int
        Index of the variable at the start of the path.
    to_index : int
        Index of the variable at the end of the path.
    min_causal_effect : float, optional (default=0.0)
        Threshold for detecting causal direction.
        Causal directions with absolute values of causal effects less than ``min_causal_effect`` are excluded.

    Returns
    -------
    paths : array-like, shape (n_paths)
        List of found path, where n_paths is the number of paths.
    effects : array-like, shape (n_paths)
        List of causal effect, where n_paths is the number of paths.
    """
    # Extract all edges
    edges = np.array(np.where(np.abs(np.nan_to_num(dag)) > min_causal_effect)).T

    # Aggregate edges by start point
    to_indices = []
    for i in range(dag.shape[0]):
        adj_list = edges[edges[:, 1] == i][:, 0].tolist()
        if len(adj_list) != 0:
            to_indices.append(adj_list)
        else:
            to_indices.append([])

    # DFS
    paths = []
    stack = [from_index]
    stack_to_indice = [to_indices[from_index]]
    while stack:
        if len(stack) > dag.shape[0]:
            raise ValueError(
                "Unable to find the path because a cyclic graph has been specified."
            )

        cur_index = stack[-1]
        to_indice = stack_to_indice[-1]

        if cur_index == to_index:
            paths.append(stack.copy())
            stack.pop()
            stack_to_indice.pop()
        else:
            if len(to_indice) > 0:
                next_index = to_indice.pop(0)
                stack.append(next_index)
                stack_to_indice.append(to_indices[next_index].copy())
            else:
                stack.pop()
                stack_to_indice.pop()

    # Calculate the causal effect for each path
    effects = []
    for p in paths:
        coefs = [dag[p[i + 1], p[i]] for i in range(len(p) - 1)]
        effects.append(np.cumprod(coefs)[-1])

    return paths, effects


def likelihood_i(x, i, b_i, bi_0):
    """Compute local log-likelihood of component i.

    Parameters
    ----------
    x : array-like, shape (n_features, n_samples)
        Data, where ``n_samples`` is the number of samples
        and ``n_features`` is the number of features.
    i : array-like
        Variable index.
    b_i : array-like
        The i^th column of adjacency matrix, B[i].
    bi_0 : float
        Constant value for the i^th variable.

    Return
    -------
    ll : float
        Local log-likelihood of component i.
    """

    sample_size = x.shape[1]  # number of data points
    var_i = variance_i(x, i, b_i)
    #
    ll = 0.0
    ll += np.sum(log_p_super_gaussian((x[i] - np.dot(b_i, x) - bi_0) / np.sqrt(var_i)))

    ll -= sample_size * np.log(np.sqrt(var_i))

    return ll


def log_p_super_gaussian(s):
    """Compute density function of the normalized independent components.

    Parameters
    ----------
    s : array-like, shape (1, n_samples)
        Data, where ``n_samples`` is the number of samples.

    Return
    -------
    x : float
        Density function of the normalized independent components,
        whose disturbances are super-Gaussian.
    """
    const = -0.35  # normalising constant
    return -np.sqrt(2.0) * np.absolute(s) + const


def variance_i(X, i, b_i):
    """Compute empirical variance of component i.

    Parameters
    ----------
    x : array-like, shape (n_features, n_samples)
        Data, where ``n_samples`` is the number of samples
        and ``n_features`` is the number of features.
    i : array-like
        Variable index.
    b_i : array-like
        The i^th column of adjacency matrix, B[i].

    Return
    -------
    variance : float
        Empirical variance of component i.
    """
    # T = X.shape[1]  # sample size
    estimated_disturbance = X[i] - np.dot(b_i, X)
    # variance = np.sum(estimated_disturbance ** 2) / T  # JMLR paper assumes zero mean
    variance = np.var(estimated_disturbance)  # stable version, even not zero mean.

    return variance



import os

import numpy as np
import pandas as pd
from lingam.bottom_up_parce_lingam import BottomUpParceLiNGAM
from sklearn.linear_model import LinearRegression


def test_fit_success():
    # causal direction: x5 --> x1, x3 --> x0 --> x2 --> x4
    n_samples = 300
    x5 = np.random.uniform(size=n_samples)
    x1 = 2.0 * x5 + np.random.uniform(size=n_samples)
    x3 = 2.0 * x5 + np.random.uniform(size=n_samples)
    x0 = 0.5 * x1 - 0.5 * x3 + np.random.uniform(size=n_samples)
    x2 = 0.5 * x0 + np.random.uniform(size=n_samples)
    x4 = 1.0 * x2 + np.random.uniform(size=n_samples)
    # x5 is latent confounders
    X = pd.DataFrame(
        np.array([x0, x1, x2, x3, x4]).T, columns=['x0', 'x1', 'x2', 'x3', 'x4']
    )

    model = BottomUpParceLiNGAM()
    model.fit(X)

    co = model.causal_order_
    am = model.adjacency_matrix_

    te = model.estimate_total_effect(X, 0, 4)
    # assert te > 0.3 and te < 0.7

    te = model.estimate_total_effect(X, 4, 0)
    te = model.estimate_total_effect(X, 4, 2)

    p_values = model.get_error_independence_p_values(X)
    # assert p_values.shape[0] == 5 and p_values.shape[1] == 5

    # reject
    model = BottomUpParceLiNGAM(alpha=1.0)
    model.fit(X)
    te = model.estimate_total_effect(X, 4, 0)
    p_values = model.get_error_independence_p_values(X)

    # coverage
    model = BottomUpParceLiNGAM(alpha=0.0)
    model.fit(X)
    model._adjacency_matrix[0, 1] = np.nan
    model._adjacency_matrix[1, 0] = np.nan
    p_values = model.get_error_independence_p_values(X)

    model = BottomUpParceLiNGAM(alpha=0.0)
    model.fit(X)
    model._causal_order = [[0, 1], 2, 3, 4]
    model.estimate_total_effect(X, 0, 1)

def test_bootstrap_success():
    # causal direction: x5 --> x1, x3 --> x0 --> x2 --> x4
    n_samples = 300
    x5 = np.random.uniform(size=n_samples)
    x1 = 2.0 * x5 + np.random.uniform(size=n_samples)
    x3 = 2.0 * x5 + np.random.uniform(size=n_samples)
    x0 = 0.5 * x1 - 0.5 * x3 + np.random.uniform(size=n_samples)
    x2 = 0.5 * x0 + np.random.uniform(size=n_samples)
    x4 = 1.0 * x2 + np.random.uniform(size=n_samples)
    # x5 is latent confounders
    X = pd.DataFrame(
        np.array([x0, x1, x2, x3, x4]).T, columns=['x0', 'x1', 'x2', 'x3', 'x4']
    )

    model = BottomUpParceLiNGAM()
    result = model.bootstrap(X, n_sampling=5)
    am = result.adjacency_matrices_
    assert len(am) == 5
    te = result.total_effects_
    assert len(te) == 5

    cdc = result.get_causal_direction_counts()
    dagc = result.get_directed_acyclic_graph_counts()
    probs = result.get_probabilities()
    ce = result.get_total_causal_effects()
    paths = result.get_paths(3, 4)

def test_fit_invalid_data():
    # Not array data
    X = 1
    try:
        model = BottomUpParceLiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include non-numeric data
    x0 = np.random.uniform(size=5)
    x1 = np.array(['X', 'Y', 'X', 'Y', 'X'])
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    try:
        model = BottomUpParceLiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include NaN values
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    X.iloc[100, 0] = np.nan
    try:
        model = BottomUpParceLiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include infinite values
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    X.iloc[100, 0] = np.inf
    try:
        model = BottomUpParceLiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # invalid regressor
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    try:
        model = BottomUpParceLiNGAM(regressor=1)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # invalid alpha
    try:
        model = BottomUpParceLiNGAM(alpha=-1)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError


def test_prior_knowledge_success():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=300)
    x1 = 2.0 * x0 + np.random.uniform(size=300)
    x2 = np.random.uniform(size=300)
    x3 = 4.0 * x1 + np.random.uniform(size=300)
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=['x0', 'x1', 'x2', 'x3'])

    # prior knowledge: nothing
    pk = np.array(
        [
            [0, -1, -1, -1],
            [-1, 0, -1, -1],
            [-1, -1, 0, -1],
            [-1, -1, -1, 0],
        ]
    )

    model = BottomUpParceLiNGAM(prior_knowledge=pk)
    model.fit(X)
    co = model.causal_order_
    print(co)

    # prior knowledge: x1 is exogenous
    pk = np.array(
        [
            [0, -1, -1, -1],
            [0, 0, 0, 0],
            [-1, -1, 0, -1],
            [-1, -1, -1, 0],
        ]
    )

    model = BottomUpParceLiNGAM(prior_knowledge=pk)
    model.fit(X)
    co = model.causal_order_
    print(co)

    # prior knowledge: x0 is sink
    pk = np.array(
        [
            [0, -1, -1, -1],
            [0, 0, -1, -1],
            [0, -1, 0, -1],
            [0, -1, -1, 0],
        ]
    )

    model = BottomUpParceLiNGAM(prior_knowledge=pk)
    model.fit(X)
    co = model.causal_order_
    print(co)

    # prior knowledge: x2-->x3 has path
    pk = np.array(
        [
            [0, -1, -1, -1],
            [-1, 0, -1, -1],
            [-1, -1, 0, 0],
            [-1, -1, 1, 0],
        ]
    )

    model = BottomUpParceLiNGAM(prior_knowledge=pk)
    model.fit(X)
    co = model.causal_order_
    print(co)

    # prior knowledge: x1-->x3 does not have path
    pk = np.array(
        [
            [0, -1, -1, -1],
            [-1, 0, -1, -1],
            [-1, -1, 0, -1],
            [-1, 0, -1, 0],
        ]
    )

    model = BottomUpParceLiNGAM(prior_knowledge=pk)
    model.fit(X)


def test_prior_knowledge_invalid():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0 * x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=['x0', 'x1', 'x2', 'x3'])

    # prior knowledge: invalid
    pk = np.array(
        [
            [0, -1, 1],
            [-1, 0, -1],
            [-1, -1, 0],
            [-1, -1, -1],
        ]
    )

    try:
        model = BottomUpParceLiNGAM(prior_knowledge=pk)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # prior knowledge: inconsistent
    pk = np.array(
        [
            [0, 1, -1, -1],
            [1, 0, -1, -1],
            [-1, -1, 0, -1],
            [-1, -1, -1, 0],
        ]
    )

    try:
        model = BottomUpParceLiNGAM(prior_knowledge=pk)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError


def test_fit_regressor():
    # causal direction: x5 --> x1, x3 --> x0 --> x2 --> x4
    n_samples = 300
    x5 = np.random.uniform(size=n_samples)
    x1 = 2.0 * x5 + np.random.uniform(size=n_samples)
    x3 = 2.0 * x5 + np.random.uniform(size=n_samples)
    x0 = 0.5 * x1 - 0.5 * x3 + np.random.uniform(size=n_samples)
    x2 = 0.5 * x0 + np.random.uniform(size=n_samples)
    x4 = 1.0 * x2 + np.random.uniform(size=n_samples)
    # x5 is latent confounders
    X = pd.DataFrame(
        np.array([x0, x1, x2, x3, x4]).T, columns=['x0', 'x1', 'x2', 'x3', 'x4']
    )

    reg = LinearRegression()

    model = BottomUpParceLiNGAM(regressor=reg)
    model.fit(X)


def test_bootstrap_invalid_data():
    x0 = np.random.uniform(size=300)
    x1 = np.random.uniform(size=300)
    x2 = np.random.uniform(size=300)
    x3 = np.random.uniform(size=300)
    X_success = pd.DataFrame(
        np.array([x0, x1, x2, x3]).T, columns=['x0', 'x1', 'x2', 'x3']
    )

    # Invalid argument: bootstrap(n_sampling=-1)
    model = BottomUpParceLiNGAM()
    try:
        result = model.bootstrap(X_success, n_sampling=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: bootstrap(n_sampling='3')
    model = BottomUpParceLiNGAM()
    try:
        result = model.bootstrap(X_success, n_sampling='3')
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(n_directions=-1)
    model = BottomUpParceLiNGAM()
    result = model.bootstrap(X_success, n_sampling=5)
    try:
        result.get_causal_direction_counts(n_directions=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(n_directions='3')
    model = BottomUpParceLiNGAM()
    result = model.bootstrap(X_success, n_sampling=5)
    try:
        result.get_causal_direction_counts(n_directions='3')
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(min_causal_effect=-1.0)
    try:
        result.get_causal_direction_counts(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(n_dags=-1)
    try:
        result.get_directed_acyclic_graph_counts(n_dags=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(n_dags='3')
    try:
        result.get_directed_acyclic_graph_counts(n_dags='3')
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(min_causal_effect=-1.0)
    try:
        result.get_directed_acyclic_graph_counts(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_probabilities(min_causal_effect=-1.0)
    try:
        result.get_probabilities(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_total_causal_effects(min_causal_effect=-1.0)
    try:
        result.get_total_causal_effects(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_paths(min_causal_effect=-1.0)
    try:
        result.get_paths(0, 1, min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError



import os

import numpy as np
import pandas as pd

from lingam.multi_group_direct_lingam import MultiGroupDirectLiNGAM


def test_fit_success():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=500)
    x1 = 2.1*x0 + np.random.uniform(size=500)
    x2 = np.random.uniform(size=500)
    x3 = 4.2*x1 + np.random.uniform(size=500)
    X2 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1, X2]

    model = MultiGroupDirectLiNGAM()
    model.fit(X_list)

    # check the causal ordering
    co = model.causal_order_
    assert co.index(0) < co.index(1) < co.index(3)

    # check the adjacency matrix
    am = model.adjacency_matrices_[0]
    # assert am[1, 0] > 1.5 and am[3, 1] > 3.5

    # check the adjacency matrix
    am = model.adjacency_matrices_[1]
    # assert am[1, 0] > 1.5 and am[3, 1] > 3.5

    am[1, 0] = 0.0
    am[3, 1] = 0.0
    # assert np.sum(am) < 0.1

    co = model.causal_order_
    am = model.adjacency_matrices_
    p_values = model.get_error_independence_p_values(X_list)


def test_fit_invalid_data():
    # Different features
    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000)
    x1 = 2.1*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    X2 = pd.DataFrame(np.array([x0, x1, x2]).T,
                      columns=['x0', 'x1', 'x2'])

    X_list = [X1, X2]

    try:
        model = MultiGroupDirectLiNGAM()
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Not list data
    X = 1
    try:
        model = MultiGroupDirectLiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include not-array data
    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1, 1]

    try:
        model = MultiGroupDirectLiNGAM()
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

    # 
    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1]

    try:
        model = MultiGroupDirectLiNGAM()
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include non-numeric data
    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=5)
    x1 = np.array(['X', 'Y', 'X', 'Y', 'X'])
    X2 = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])

    X_list = [X1, X2]

    try:
        model = MultiGroupDirectLiNGAM()
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include NaN values
    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    X2 = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    X2.iloc[100, 0] = np.nan

    X_list = [X1, X2]

    try:
        model = MultiGroupDirectLiNGAM()
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include infinite values
    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    X.iloc[100, 0] = np.inf

    X_list = [X1, X2]

    try:
        model = MultiGroupDirectLiNGAM()
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError


def test_prior_knowledge_success():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000)
    x1 = 2.1*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.2*x1 + np.random.uniform(size=1000)
    X2 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1, X2]

    # prior knowledge: nothing
    pk = np.array([
        [0, -1, -1, -1],
        [-1,  0, -1, -1],
        [-1, -1,  0, -1],
        [-1, -1, -1,  0],
    ])

    model = MultiGroupDirectLiNGAM(prior_knowledge=pk)
    model.fit(X_list)
    co = model.causal_order_
    assert co.index(0) < co.index(1) < co.index(3)

    # prior knowledge: x1 is exogenous
    pk = np.array([
        [0, -1, -1, -1],
        [0,  0,  0,  0],
        [-1, -1,  0, -1],
        [-1, -1, -1,  0],
    ])

    model = MultiGroupDirectLiNGAM(prior_knowledge=pk)
    model.fit(X_list)
    co = model.causal_order_
    print(co)
    assert co.index(1) < co.index(0)
    assert co.index(1) < co.index(2)
    assert co.index(1) < co.index(3)

    # prior knowledge: x0 is sink
    pk = np.array([
        [0, -1, -1, -1],
        [0,  0, -1, -1],
        [0, -1,  0, -1],
        [0, -1, -1,  0],
    ])

    model = MultiGroupDirectLiNGAM(prior_knowledge=pk)
    model.fit(X_list)
    co = model.causal_order_
    print(co)
    assert co.index(0) > co.index(1)
    assert co.index(0) > co.index(2)
    assert co.index(0) > co.index(3)

    # prior knowledge: x2-->x3 has path
    pk = np.array([
        [0, -1, -1, -1],
        [-1,  0, -1, -1],
        [-1, -1,  0,  0],
        [-1, -1,  1,  0],
    ])

    model = MultiGroupDirectLiNGAM(prior_knowledge=pk)
    model.fit(X_list)
    co = model.causal_order_
    print(co)
    assert co.index(2) < co.index(3)

    # prior knowledge: x1-->x3 does not have path
    pk = np.array([
        [0, -1, -1, -1],
        [-1,  0, -1, -1],
        [-1, -1,  0, -1],
        [-1,  0, -1,  0],
    ])

    model = MultiGroupDirectLiNGAM(prior_knowledge=pk)
    model.fit(X_list)


def test_prior_knowledge_invalid():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000)
    x1 = 2.1*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.2*x1 + np.random.uniform(size=1000)
    X2 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1, X2]

    # prior knowledge: invalid
    pk = np.array([
        [0, -1, 1],
        [-1,  0, -1],
        [-1, -1,  0],
        [-1, -1, -1],
    ])

    try:
        model = MultiGroupDirectLiNGAM(prior_knowledge=pk)
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

def test_bootstrap_success():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000)
    x1 = 2.1*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.2*x1 + np.random.uniform(size=1000)
    X2 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1, X2]

    model = MultiGroupDirectLiNGAM()
    result_list = model.bootstrap(X_list, n_sampling=5)

    result = result_list[0]
    result.adjacency_matrices_
    result.total_effects_

    # No argument
    cdc = result.get_causal_direction_counts()

    # n_directions=2
    cdc = result.get_causal_direction_counts(n_directions=2)

    # min_causal_effect=0.2
    cdc = result.get_causal_direction_counts(min_causal_effect=0.2)

    # split_by_causal_effect_sign=True
    cdc = result.get_causal_direction_counts(split_by_causal_effect_sign=True)

    # No argument
    dagc = result.get_directed_acyclic_graph_counts()

    # n_dags=2
    dagc = result.get_directed_acyclic_graph_counts(n_dags=2)

    # min_causal_effect=0.6
    dagc = result.get_directed_acyclic_graph_counts(min_causal_effect=0.6)

    # split_by_causal_effect_sign=True
    dagc = result.get_directed_acyclic_graph_counts(split_by_causal_effect_sign=True)

    # get_probabilities
    probs = result.get_probabilities()

    # get_probabilities
    probs = result.get_probabilities(min_causal_effect=0.6)

    # get_total_causal_effects
    ce = result.get_total_causal_effects()

    # get_total_causal_effects
    ce = result.get_total_causal_effects(min_causal_effect=0.6)

def test_bootstrap_invalid_data():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000)
    x1 = 2.1*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.2*x1 + np.random.uniform(size=1000)
    X2 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1, X2]

    # Invalid argument: bootstrap(n_sampling=-1)
    model = MultiGroupDirectLiNGAM()
    try:
        result = model.bootstrap(X_list, n_sampling=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: bootstrap(n_sampling='3')
    model = MultiGroupDirectLiNGAM()
    try:
        result = model.bootstrap(X_list, n_sampling='3')
    except ValueError:
        pass
    else:
        raise AssertionError

def test_estimate_total_effect_invalid():
    x0 = np.random.uniform(size=1000)
    x1 = 2.0*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0*x1 + np.random.uniform(size=1000)
    X1 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    x0 = np.random.uniform(size=1000)
    x1 = 2.1*x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.2*x1 + np.random.uniform(size=1000)
    X2 = pd.DataFrame(np.array([x0, x1, x2, x3]).T,
                      columns=['x0', 'x1', 'x2', 'x3'])

    X_list = [X1, X2]

    model = MultiGroupDirectLiNGAM()
    model.fit(X_list)
    model._causal_order = [0, 1, 2, 3]

    # warning
    model.estimate_total_effect(X_list, 3, 1)



import numpy as np
import pandas as pd

from sklearn.base import is_regressor, is_classifier
from sklearn.utils import check_array, check_random_state
from sklearn.preprocessing import OneHotEncoder
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from sklearn.linear_model import LinearRegression, LogisticRegression


class CausalDataGenerator:
    """CausalDataGenerator"""

    def __init__(self, random_state=None):
        """Construct a CausalDataGenerator.

        Parameters
        ----------
        random_state : None, int or instance of RandomState, default=None


        Attributes
        ----------
        model_cause_ : estimator object of scikit-learn
            Trained model of cause. Access will be available after fitting.

        model_effect_ : estimator object of scikit-learn
            Trained model of effect. Access will be available after fitting.

        error_cause_ : pandas.Series
            Error in the cause model. Access will be available after fitting.

        error_effect_ : pandas.Series
            Error in the effect model. Access will be available after fitting.
        """
        random_state = check_random_state(random_state)
        self._random_state = random_state

    def fit(
        self, X, cause, effect, adjustments=None, model_cause=None, model_effect=None
    ):
        """Fit models to the data.

        Parameters
        ----------
        X : pandas.DataFrame (n_samples, n_features)
            The input data. Columns with dtype of "category" are treated as categorical variables.

        cause : str
            Name of the variable to be treated as "cause". cause shall be an element of X.columns.

        effect : str
            Name of the variable to be treated as "effect". effect shall be an element of X.columns.

        adjustments : array-like of str, optional, default=None
            Names of variables to be treated as "adjustment". Each element of adjustments
            shall be an element of X.columns.

        model_cause : estimator object of scikit-learn, optional, default=None
            Model of "cause". A regression model must be set up when the objective variable is
            a continuous variable, and a classification model must be set up when the objective
            variable is a discrete variable.

        model_effect : estimator object of scikit-learn, optional, default=None
            Model of "effect". A regression model must be set up when the objective variable is
            a continuous variable, and a classification model must be set up when the objective
            variable is a discrete variable.

        Returns
        -------
        self : object
            Instance of fitted estimator.
        """

        if not isinstance(X, pd.DataFrame):
            raise TypeError("X shall be a pandas.DataFrame.")

        check_array(X, dtype=None, ensure_min_features=2)
        X = X.copy()

        if cause not in X.columns:
            raise ValueError("cause isn't exist in X.columns.")

        if effect not in X.columns:
            raise ValueError("effect isn't exist in X.columns.")

        if adjustments is None:
            adjustments = np.array([])
        else:
            adjustments = check_array(
                adjustments, ensure_2d=False, dtype=None, ensure_min_samples=0
            ).reshape((-1,))

        if model_cause is not None:
            is_categorical = X[cause].dtype == "category"
            self._is_sklearn_estimator(model_cause, is_categorical)

        if model_effect is not None:
            is_categorical = X[effect].dtype == "category"
            self._is_sklearn_estimator(model_effect, is_categorical)

        # cause
        if len(adjustments) == 0:
            model_cause, resid_cause = None, X[cause]
        else:
            model_cause, resid_cause = self._make_model(
                pd.DataFrame(X[adjustments]), X[cause], "model_cause", model=model_cause
            )

        # effect
        model_effect, resid_effect = self._make_model(
            pd.DataFrame(X[[cause, *adjustments]]),
            X[effect],
            "model_effect",
            model=model_effect,
        )

        # arguments
        self._X = X
        self._cause = cause
        self._effect = effect
        self._adjustments = adjustments

        # estimated values
        self.model_cause_ = model_cause
        self.model_effect_ = model_effect
        self.error_cause_ = resid_cause
        self.error_effect_ = resid_effect

        return self

    def generate(
        self,
        interv_endog=None,
        interv_exog=None,
        cause_model=None,
        cause_model_args=None,
        effect_model=None,
        effect_model_args=None,
    ):
        """
        Parameters
        ----------
        interv_endog : dict, optional (default=None)
            The key is the name of the target variable. The data is generated by replacing
            the specified endogenous variable with the value.
            {'name': str, 'values': array-like}.

        interv_exog : array_like, optional (default=None)
            The key is the name of the target variable. The data is generated by replacing
            the specified exogenous variable with the value.
            {'name': str, 'values': array-like}.

        cause_model : callable, optional (default=None)
            The estimated model is replaced with ``cause_model``. The arguments are
            (X, error, cause_model_args). ``X`` and ``error` are a pandas.DataFrame.
            The output of ``cause_model`` shall be an array-like and its length shall be
            the same as input data.
            ``error`` is residuals of ``cause``.

        cause_model_args : object, optional (default=None)
            Arguments of ``cause_model``.

        effect_model : callable, optional (default=None)
            Replaces the model that generates ``effect``.

        effect_model_args : object, optional (default=None)
            Arguments of ``effect_model``.

        Return
        ------
        generated_values : pd.Series
        """

        # check arguments
        interv_endog = self._check_data_dict(interv_endog, "interv_endog")
        interv_exog = self._check_data_dict(interv_exog, "interv_exog")

        if cause_model is not None:
            if not callable(cause_model):
                raise TypeError("cause_model shall be callable.")

        if effect_model is not None:
            if not callable(effect_model):
                raise TypeError("effect_model shall be callable.")

        # initial data
        generated = self._X[[self._cause, self._effect, *self._adjustments]].copy()

        # update errors
        generated[self._cause] = self.error_cause_
        generated[self._effect] = self.error_effect_

        # update interventions
        for name, interv_exog_ in interv_exog.items():
            generated[name] = interv_exog_

        # cause
        if self._cause in interv_endog.keys():
            generated[self._cause] = interv_endog[self._cause]
        else:
            if cause_model is not None:
                try:
                    predicted = cause_model(
                        generated[self._adjustments],
                        generated[self._cause],
                        cause_model_args,
                    )
                except Exception as e:
                    raise RuntimeError("Exception: cause_model: " + str(e))

                generated[self._cause] = predicted
                if self._X[self._cause].dtype == "category":
                    generated[self._cause] = pd.Categorical(generated[self._cause])
            else:
                if self._X[self._cause].dtype != "category":
                    predicted = self.model_cause_.predict(generated[self._adjustments])
                    generated[self._cause] += np.array(predicted)
                else:
                    proba = self.model_cause_.predict_proba(
                        generated[self._adjustments]
                    )

                    predicted = []
                    for proba_ in proba:
                        predicted.append(
                            self._random_state.choice(
                                self.model_cause_.classes_, p=proba_
                            )
                        )
                    generated[self._cause] = np.array(predicted)
                    generated[self._cause] = pd.Categorical(generated[self._cause])

        # effect
        if self._effect in interv_endog.keys():
            generated[self._effect] = interv_endog[self._effect]
        else:
            if effect_model is not None:
                try:
                    predicted = effect_model(
                        generated[[self._cause, *self._adjustments]],
                        generated[self._effect],
                        effect_model_args,
                    )
                except Exception as e:
                    raise RuntimeError("Exception: effect_model: " + str(e))

                generated[self._effect] = predicted
                if self._X[self._effect].dtype == "category":
                    generated[self._effect] = pd.Categorical(generated[self._effect])
            else:
                if self._X[self._effect].dtype != "category":
                    predicted = self.model_effect_.predict(
                        generated[[self._cause, *self._adjustments]]
                    )
                    generated[self._effect] += np.array(predicted)
                else:
                    proba = self.model_effect_.predict_proba(
                        generated[[self._cause, *self._adjustments]]
                    )

                    predicted = []
                    for proba_ in proba:
                        predicted.append(
                            self._random_state.choice(
                                self.model_effect_.classes_, p=proba_
                            )
                        )
                    generated[self._effect] = np.array(predicted)
                    generated[self._effect] = pd.Categorical(generated[self._effect])

        generated_values = generated.loc[self._X.index, self._X.columns]

        return generated_values

    def _is_sklearn_estimator(self, estimator, is_categorical):
        if is_categorical is True and is_regressor(estimator):
            raise TypeError(
                "The Object variable is categorical but the estimator is a regressor."
            )
        elif is_categorical is False and is_classifier(estimator):
            raise TypeError(
                "The Object variable is not categorical but the estimator is a classifier."
            )

        if is_categorical is True:
            try:
                func = getattr(estimator, "predict_proba")
                if not callable(func):
                    raise Exception
            except Exception:
                raise RuntimeError(
                    "Classification models shall have " + "predict_proba()."
                )

    def _make_model(self, X, y, name, model=None):
        if model is not None:
            try:
                model.fit(X, y)
            except Exception as e:
                raise RuntimeError(f"{name}.fit(): {str(e)}")
        else:
            if y.dtypes == "category":
                model = LogisticRegression()
            else:
                model = LinearRegression()

            categoricals = X.columns[X.dtypes == "category"]
            numerics = X.columns[X.dtypes != "category"]

            if len(categoricals) > 0:
                transformers = [
                    ("categorical", OneHotEncoder(sparse=False), categoricals),
                    ("numeric", "passthrough", numerics),
                ]
                trans = ColumnTransformer(transformers=transformers)

                model = Pipeline([("transformer", trans), ("estimator", model)])

            model.fit(X, y)

        if y.dtype == "category":
            resids = np.full(X.shape[0], np.nan)
        else:
            resids = y - model.predict(X)
        resids = pd.Series(resids, index=y.index, name=y.name)

        return model, resids

    def _check_data_dict(self, data_dict, name):
        if data_dict is None:
            return {}

        if not isinstance(data_dict, dict):
            raise TypeError(f"{name} shall be a dictionary.")

        for name, values in data_dict.items():
            if not isinstance(name, str):
                raise TypeError(f"A key of {name} shall be str.")

            if name not in [self._cause, self._effect, *self._adjustments]:
                raise ValueError(
                    f"Keys of {name} shall be cause or effect or a element of adjustments."
                )

            values = check_array(values, ensure_2d=False, dtype=None)
            if values.shape[0] != self._X.shape[0]:
                raise ValueError(f"shape[0] of {name} shall be the same as X.shape[0].")

        return data_dict



import os

import numpy as np
import pandas as pd
from lingam.resit import RESIT
from sklearn.ensemble import RandomForestRegressor


def test_fit_success():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0 * x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=["x0", "x1", "x2", "x3"])

    reg = RandomForestRegressor(max_depth=4, random_state=0)
    model = RESIT(regressor=reg)
    model.fit(X)

    # check the causal ordering
    co = model.causal_order_

    # check the adjacency matrix
    am = model.adjacency_matrix_

    # check estimate_total_effect (Not implement)
    te = model.estimate_total_effect(X, 0, 3)

    # check get_error_independence_p_values (Not implement)
    p_values = model.get_error_independence_p_values(X)


def test_fit_invalid():
    reg = RandomForestRegressor(max_depth=4, random_state=0)

    # Not array data
    X = 1
    try:
        model = RESIT(regressor=reg)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0 * x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=["x0", "x1", "x2", "x3"])

    # Invalid regressor
    try:
        model = RESIT(regressor=None)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid regressor
    dummy_reg = lambda x: x
    try:
        model = RESIT(regressor=dummy_reg)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid alpha
    try:
        model = RESIT(regressor=reg, alpha=-1)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError
